"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=duplicate-code
from __future__ import annotations
from functools import cached_property
from uuid import uuid4
from pydantic.dataclasses import dataclass


from loom.eka import (
    Block,
    Lattice,
    LatticeType,
    PauliOperator,
    Stabilizer,
    SyndromeCircuit,
    Circuit,
    Channel,
)
from loom.eka.utilities import dataclass_config


@dataclass(config=dataclass_config)
class SteaneCode(Block):
    """
    A sub-class of ``Block`` that represents a Steane code block.
    """

    @classmethod
    def create(  # pylint: disable=too-many-locals
        cls,
        lattice: Lattice,
        unique_label: str = None,
        position: tuple[int, ...] = (0, 0),
        skip_validation: bool = False,
    ) -> SteaneCode:
        """Create Steane Code Block object. The Steane Code is a
        [[7, 1, 3]] quantum error-correcting CSS code. There are 6 stabilizers, three
        of X type and three of Z type all of support size 4. We implement a naive
        syndrome extraction scheme, where measure all X stabilizers first and all the
        Z stabilizers later. We choose to embed the code in a 2D square lattice, where
        the logical operators are defined on the three lowest data qubits.

        Parameters
        ----------
        lattice : Lattice
            Lattice on which the block is defined. The qubit indices depend on the type
            of lattice.
        unique_label : str, optional
            Label for the block. It must be unique among all blocks in the initial CRD.
            If no label is provided, a unique label is generated automatically using the
            uuid module.
        position : tuple[int, ...], optional
            Position of the top left corner of the block on the lattice,
            by default (0, 0).
        skip_validation : bool, optional
            Skip validation of the block object, by default False.

        Returns
        -------
        Block
            Block object for the Steane Code.
        """

        # Input validation
        if lattice.lattice_type != LatticeType.SQUARE_2D:
            raise ValueError(
                "The creation of Steane code blocks is "
                "currently only supported for 2D square lattices. Instead "
                f"the lattice is of type {lattice.lattice_type}."
            )

        if not isinstance(position, tuple) or any(
            not isinstance(x, int) for x in position
        ):
            raise ValueError(
                f"`position` must be a tuple of integers. Got '{position}' instead."
            )

        if unique_label is None:
            unique_label = str(uuid4())

        # Define the stabilizers
        stabilizer_supports = [
            [(0, 0, 0), (0, 1, 0), (1, 1, 0), (1, 0, 0)],
            [(1, 0, 0), (1, 1, 0), (2, 1, 0), (2, 0, 0)],
            [(2, 1, 0), (1, 2, 0), (0, 1, 0), (1, 1, 0)],
        ]

        x_stabilizers = [
            Stabilizer(pauli="X" * 4, data_qubits=supp, ancilla_qubits=[(i, 0, 1)])
            for i, supp in enumerate(stabilizer_supports)
        ]
        z_stabilizers = [
            Stabilizer(pauli="Z" * 4, data_qubits=supp, ancilla_qubits=[(0, i + 1, 1)])
            for i, supp in enumerate(stabilizer_supports)
        ]

        stabilizers = x_stabilizers + z_stabilizers

        # Define the logical operators
        logical_data_qubits = [(i, 0, 0) for i in range(3)]

        logical_x_operator = PauliOperator(
            pauli="XXX",
            data_qubits=logical_data_qubits,
        )
        logical_z_operator = PauliOperator(
            pauli="ZZZ",
            data_qubits=logical_data_qubits,
        )

        # Define the syndrome extraction circuits
        x_syndrome_circuits = cls.generate_syndrome_extraction_circuits("XXXX")
        z_syndrome_circuits = cls.generate_syndrome_extraction_circuits("ZZZZ")
        syndrome_circuits = [x_syndrome_circuits, z_syndrome_circuits]

        # Define the stabilizer to circuit mapping
        stabilizer_to_circuit = {
            stab.uuid: x_syndrome_circuits.uuid for stab in x_stabilizers
        } | {stab.uuid: z_syndrome_circuits.uuid for stab in z_stabilizers}

        block = cls(
            unique_label=unique_label,
            stabilizers=stabilizers,
            logical_x_operators=[logical_x_operator],
            logical_z_operators=[logical_z_operator],
            syndrome_circuits=syndrome_circuits,
            stabilizer_to_circuit=stabilizer_to_circuit,
            skip_validation=skip_validation,
        )

        if position == (0, 0):
            return block

        return block.shift(position)

    @staticmethod
    def generate_syndrome_extraction_circuits(pauli: str) -> SyndromeCircuit:
        """
        Generate syndrome extraction circuit for a stabilizers from the Steane Code.
        The prescription followed here is to first measure all stabilizers of a given
        type simultaneously in four layers (arbitrarily chosen to be entangling in the
        data qubits with the ancilla in a clockwise order). We choose X stabilizers to
        be measured first, thus we need to add four idling steps at the end, and four
        at the beginning for the Z stabilizers. The ancilla is then measured and reset.

        Parameters
        ----------
        pauli: str
            Pauli operator for which the syndrome extraction circuit is generated.

        Returns
        -------
        SyndromeCircuit
            Syndrome extraction circuit for the given Pauli operator.
        """

        # Extract parameters
        name = f"{pauli}_syndrome_extraction"
        pauli_type = pauli[0]

        # Define channels
        data_channels = [Channel(type="quantum", label=f"d{i}") for i in range(4)]
        cbit_channel = Channel(type="classical", label="c0")
        ancilla_channel = Channel(type="quantum", label="a0")

        # Define Hadamard gates
        hadamard1 = tuple([Circuit("H", channels=[ancilla_channel])])
        hadamard2 = tuple([Circuit("H", channels=[ancilla_channel])])

        # Entangling layer
        entangle_ancilla = [
            [Circuit(f"C{p}", channels=[ancilla_channel, data_channels[i]])]
            for i, p in enumerate(pauli)
        ]

        # Add idling step
        if pauli_type == "Z":
            entangle_ancilla = [(), (), (), ()] + entangle_ancilla
        else:
            entangle_ancilla += [(), (), (), ()]

        # Add ancilla measurement and reset
        measurement = tuple(
            [Circuit("Measurement", channels=[ancilla_channel, cbit_channel])]
        )
        reset = tuple([Circuit("Reset_0", channels=[ancilla_channel])])

        # Compose circuit operations as a list
        circuit_list = [reset, hadamard1] + entangle_ancilla + [hadamard2, measurement]

        # Return the syndrome extraction circuit
        return SyndromeCircuit(
            pauli=pauli,
            name=name,
            circuit=Circuit(
                name=name,
                circuit=circuit_list,
                channels=data_channels + [ancilla_channel, cbit_channel],
            ),
        )

    # Instance methods
    def __eq__(self, other: SteaneCode) -> bool:
        if not isinstance(other, SteaneCode):
            raise NotImplementedError(f"Cannot compare SteaneCode with {type(other)}")
        return super().__eq__(other)

    def shift(
        self, position: tuple[int, ...], new_label: str | None = None
    ) -> SteaneCode:
        return super().shift(position, new_label)

    def rename(self, name: str) -> SteaneCode:
        return super().rename(name)

    @cached_property
    def stabilizers_labels(self) -> dict[str, dict[str, tuple[int, ...]]]:
        return super().stabilizers_labels



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from pydantic import Field
from pydantic.dataclasses import dataclass

from .check_code_stabilizers import CodeStabilizerCheck
from .check_logical_ops import LogicalOperatorCheck
from .check_stabilizer_measurement import StabilizerMeasurementCheck


@dataclass(frozen=True)
class AllChecks:
    """
    Dataclass to store the results of all checks.

    Parameters
    ----------
    CodeStabilizers: CodeStabilizerCheck
        The results of the Code Stabilizer check.
    LogicalState: LogicalOperatorCheck
        The results of the Logical Operator check.
    StabilizersMeasured: StabilizerMeasurementCheck
        The results of the Stabilizer Measurement check.
    """

    code_stabilizers: CodeStabilizerCheck
    logical_operators: LogicalOperatorCheck
    stabilizers_measured: StabilizerMeasurementCheck

    def __iter__(self):
        """Iterate over the checks."""
        return iter(self.__dict__.values())


@dataclass(frozen=True)
class DebugData:
    """Dataclass to store the results of validator verification checks.

    Parameters
    ----------
    checks: AllChecks
        The results of all checks.
    valid: bool
        True if all checks have passed, False otherwise.
    """

    checks: AllChecks
    valid: bool = Field(init_var=False, default=True)

    def __post_init__(self):
        # Check if all checks are valid and set the valid attribute
        valid = all(check.valid for check in self.checks)
        object.__setattr__(self, "valid", valid)

    def __str__(self):
        """String representation of the DebugData."""
        out = f"DebugData: valid = {self.valid}\n"
        for check in self.checks:
            if not check.valid:
                # Add a separator for better readability
                out += "-" * 40 + "\n"
                out += str(check) + "\n"
        return out.rstrip("\n")



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=duplicate-code
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

from .base_operation import Operation, OpType


@dataclass
class ControlledOperation(Operation):
    """
    A decorator of the Operation class that turns any operation into a classically
    controlled operation.
    The Operation will be conditioned on the value of a single bit.

    NOTE: The Operation classes decorated by the ControlledOperation class can only
    be controlled by one bit.

    Parameters
    ----------
    app_operation: :class:`loom.cliffordsim.operations.base_operation.Operation`
        The Operation that will be Classically Controlled.
    reg_name: str
        The name of the Classical Register where the bit that classically controls
        app_operation will be located.
    bit_order: Optional[int]
        The ordering of the bit within the classical register that app_operation will
        be conditioned on. Either bit_order or bit_id needs to be specified. If both
        or none are specified, an error is raised.
    bit_id: Optional[str]
        The bit ID of the bit within the classical register that app_operation will be
        conditioned on. Either bit_order or bit_id needs to be specified. If both or
        none are specified, an error is raised.
    """

    operation_type: Enum = field(default=OpType.CCONTROL, init=False)
    name: str = field(default="ControlledOperation", init=False)
    app_operation: Operation
    reg_name: str
    bit_order: Optional[int] = field(default=None)
    bit_id: Optional[str] = field(default=None)

    def __post_init__(self):
        if self.bit_order is None and self.bit_id is None:
            raise ValueError(
                "Either the bit order or the bit id of the classical bit has to be"
                "specified."
            )
        if self.bit_order is not None and self.bit_id is not None:
            raise ValueError(
                "Both bit_order and bit_id cannot be specified together. Only input 1 "
                "parameter, not both."
            )


def has_ccontrol(cls: Operation):
    """
    This function acts as a decorator that adds the with_ccontrol method to the input
    Operation class.
    """

    def with_ccontrol(self, reg_name, bit_order=None, bit_id=None):
        """
        This method returns a ControlledOperation wrapped version of the Operation
        class. The wrapped Operation class is classically controlled by the bit from
        the classical register specified by reg_name in bit_order or by bit_id.

        NOTE: ControlledOperation(s) can only be conditioned on one classical bit.
        """
        return ControlledOperation(
            app_operation=self, reg_name=reg_name, bit_order=bit_order, bit_id=bit_id
        )

    setattr(cls, "with_ccontrol", with_ccontrol)
    return cls



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from enum import Enum
from itertools import chain
from pydantic import field_validator
from pydantic.dataclasses import dataclass

from ..eka.utilities import BoolOp


class OpType(Enum):
    """Enum for operation types."""

    SINGLE_QUBIT = "single_qubit"
    TWO_QUBIT = "two_qubit"
    MEASUREMENT = "measurement"
    RESET = "reset"
    CONTROL_FLOW = "control_flow"
    SUBCIRCUIT = "subcircuit"
    BOOL_LOGIC = "bool_logic"
    UTILS = "utils"
    CUSTOM = "custom"


@dataclass(frozen=True)
class OpSignature:
    """
    A class to represent a channel signature for quantum operations.

    Parameters
    ----------
    name : str
        The name of the channel signature.
    quantum_input : int
        The number of quantum inputs required by the operation.
    classical_input : int
        The number of classical inputs required by the operation.
    description : str, optional
        A description of the channel signature.
        Default is an empty string.
    """

    name: str
    op_type: OpType
    quantum_input: int = 0
    classical_input: int = 0
    is_clifford: bool = True
    description: str = ""

    @field_validator("quantum_input", "classical_input")
    @classmethod
    def validate_positive_inputs(cls, v):
        """Validate that quantum_input and classical_input are non-negative."""
        if v < 0:
            raise ValueError("Input count must be non-negative")
        return v

    @classmethod
    def single_qubit_op_signature(
        cls, name: str, is_clifford: bool = True
    ) -> "OpSignature":
        """Create a single qubit operation signature."""
        return cls(
            name=name,
            op_type=OpType.SINGLE_QUBIT,
            quantum_input=1,
            is_clifford=is_clifford,
        )

    @classmethod
    def reset_op_signature(cls, name: str) -> "OpSignature":
        """Create a reset operation signature."""
        return cls(
            name=name,
            op_type=OpType.RESET,
            quantum_input=1,
            is_clifford=True,
        )

    @classmethod
    def two_qubit_op_signature(cls, name: str) -> "OpSignature":
        """Create a two qubit operation signature."""
        return cls(name=name, op_type=OpType.TWO_QUBIT, quantum_input=2)

    @classmethod
    def measurement_op_signature(cls, name: str) -> "OpSignature":
        """Create a measurement operation signature."""
        return cls(
            name=name, op_type=OpType.MEASUREMENT, quantum_input=1, classical_input=1
        )


CONTROL_FLOW_OP_SIGNATURE = frozenset(
    {
        OpSignature(
            name="classical_if",
            op_type=OpType.CONTROL_FLOW,
        ),
        OpSignature(
            name="classical_else",
            op_type=OpType.CONTROL_FLOW,
        ),
        OpSignature(
            name="end_if",
            op_type=OpType.CONTROL_FLOW,
        ),
    }
)

UTILS_SIGNATURE = frozenset(
    {
        OpSignature(
            name="comment",
            op_type=OpType.UTILS,
        ),
        OpSignature(
            name="indent_more",
            op_type=OpType.UTILS,
        ),
        OpSignature(
            name="indent_less",
            op_type=OpType.UTILS,
        ),
    }
)

BOOL_LOGIC_OP_SIGNATURE = frozenset(
    {
        OpSignature(
            name=n,
            op_type=OpType.BOOL_LOGIC,
            classical_input=2,
            is_clifford=False,
        )
        for n in BoolOp.multi_bit_list()
    }
    | {
        OpSignature(
            name=n,
            op_type=OpType.BOOL_LOGIC,
            classical_input=1,
            is_clifford=False,
        )
        for n in BoolOp.mono_bit_list()
    }
)

STR_SINGLE_QUBIT_CLIFFORD_GATE = frozenset(
    {"i", "x", "y", "z", "h", "phase", "phaseinv"}
)
STR_NONCLIFFORD_SINGLE_QUBIT_GATE = frozenset({"t"})
STR_TWO_QUBIT_GATE = frozenset({"cnot", "cx", "cy", "cz", "swap"})
STR_RESET = frozenset(
    {
        "reset",
        "reset_0",
        "reset_1",
        "reset_+",
        "reset_-",
        "reset_+i",
        "reset_-i",
    }
)
STR_MEAS = frozenset({"measurement", "measure_z", "measure_x", "measure_y"})

CLIFFORD_GATES_SIGNATURE = frozenset(
    chain(
        (
            OpSignature.single_qubit_op_signature(name=g)
            for g in STR_SINGLE_QUBIT_CLIFFORD_GATE
        ),
        (OpSignature.two_qubit_op_signature(name=g) for g in STR_TWO_QUBIT_GATE),
        (OpSignature.measurement_op_signature(name=g) for g in STR_MEAS),
        (OpSignature.reset_op_signature(name=g) for g in STR_RESET),
    )
)

NONCLIFFORD_GATES_SIGNATURE = frozenset(
    OpSignature.single_qubit_op_signature(name=g, is_clifford=False)
    for g in STR_NONCLIFFORD_SINGLE_QUBIT_GATE
)
ALL_EKA_OP_SIGNATURES = frozenset(
    CONTROL_FLOW_OP_SIGNATURE
    | BOOL_LOGIC_OP_SIGNATURE
    | CLIFFORD_GATES_SIGNATURE
    | NONCLIFFORD_GATES_SIGNATURE
    | UTILS_SIGNATURE
)

USUAL_QUANTUM_GATES = frozenset(CLIFFORD_GATES_SIGNATURE)
USUAL_CLIFFORD_GATES = frozenset(CLIFFORD_GATES_SIGNATURE)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from abc import ABC, abstractmethod
from typing import List

import numpy as np

from ..operations.base_operation import Operation
from ..tableau import Tableau
from ..pauli_frame import PauliFrame
from ..data_store import DataStore
from .instruction import Instruction, IdentityInstruction, DecoratorSelector


class MomentInterface(ABC):  # pylint: disable=too-few-public-methods
    """
    An abstract base class for Moments.
    """

    @abstractmethod
    def transform_tab(self, input_tableau: np.ndarray) -> np.ndarray:
        """
        Transform the input tableau and return the transformed tableau.
        """


class Moment(MomentInterface):
    """
    A Moment contains a set of instructions that are directly applicable
    on the Tableau.
    """

    def __init__(self, input_operations: tuple[Operation], time_step: int) -> None:
        self.root_operations = input_operations
        self.time_step = time_step
        self.instruction = self._create_instruction()

    def _create_instruction(self) -> Instruction:
        """
        Maps Operations to Instructions and builds the Moment's
        Instruction as an attribute, instruction.
        """
        # Converts the first Operation in root_operations into an Instruction
        # and wraps that Instruction around IdentityInstruction
        main_inst = IdentityInstruction()
        wrapped_inst_dec = getattr(
            DecoratorSelector, self.root_operations[0].name
        ).value
        wrapped_inst = wrapped_inst_dec(main_inst, self.root_operations[0])

        # Converts subsequent Operation(s) into Instruction(s) and wrap those
        # around the previous Instruction
        for each_operation in self.root_operations[1:]:
            inst_decorator = getattr(DecoratorSelector, each_operation.name).value
            wrapped_inst = inst_decorator(wrapped_inst, each_operation)
        return wrapped_inst

    # pylint: disable=arguments-differ
    def transform_tab(
        self, input_tableau: Tableau, data_store: DataStore, **kwargs
    ) -> Tableau:
        """
        Applies tranformation of Instructions on incoming Tableau.
        Also requires a DataStore that manages information keeping throughout
        the instruction application process and kwargs to transparently support
        future extensions.
        """
        return self.instruction.apply(input_tableau, data_store, **kwargs)

    def transform_pf(
        self, input_pauliframes: List[PauliFrame], data_store: DataStore, **kwargs
    ) -> List[PauliFrame]:
        """
        Applies tranformation of Instructions on incoming PauliFrame.
        Also requires a DataStore that manages information keeping throughout
        the instruction application process and kwargs to transparently support
        future extensions.
        """
        return self.instruction.apply_pf(input_pauliframes, data_store, **kwargs)

    def transform_pf_back(
        self, input_pauliframes: List[PauliFrame], data_store: DataStore, **kwargs
    ) -> List[PauliFrame]:
        """
        Applies backwards tranformation of Instructions on incoming PauliFrame.
        Also requires a DataStore that manages information keeping throughout
        the instruction application process and kwargs to transparently support
        future extensions.
        """
        return self.instruction.apply_pf_back(input_pauliframes, data_store, **kwargs)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations
from dataclasses import dataclass
from copy import deepcopy

import numpy as np

from loom.eka.utilities import is_stabarray_equivalent, StabArray, sparse_formatter

from .exceptions import TableauSizeError


@dataclass(frozen=True)
class TableauSnapshot:
    """
    A snapshot of the Tableau at a specific point in time.
    """

    nqubits: int
    rand_gen: int
    tableau_w_scratch: np.ndarray


class Tableau:  # pylint: disable=too-many-instance-attributes
    """
    The Tableau class represents the stabilizer state of a quantum system
    using the tableau representation. It provides methods to manipulate and
    query the stabilizer state.
    """

    dtype = np.int8

    def __init__(
        self, nqubits: int = None, initial_tableau: np.ndarray = None, seed: int = None
    ) -> None:
        """
        Initializes the the state `|00...0>`.
        """
        if not (nqubits is None) ^ (initial_tableau is None):
            raise ValueError(
                "Tableau should be defined using nqubits XOR initial_tableau"
            )

        self.rand_gen = np.random.default_rng(seed)

        if nqubits is not None:
            # construct tableau array as concatenation of an identity array
            # and an empty vector
            self.tableau_w_scratch = np.zeros(
                (2 * nqubits + 1, 2 * nqubits + 1), dtype=self.dtype
            )
            id_array = np.identity(2 * nqubits, dtype=self.dtype)
            self.tableau_w_scratch[:-1, :-1] = id_array
        else:
            self._init_w_tableau(initial_tableau)

        self.update()

    def _init_w_tableau(self, input_tableau: np.ndarray):
        """
        Assigns a custom tableau to the Tableau. The method adds an empty
        scratch row to the custom tableau and assigns it to the tableau_w_scratch
        attribute.
        """
        tableau = np.array(input_tableau, dtype=self.dtype)
        # form the tableau with the scratch row
        scratch_row = np.zeros((1, tableau.shape[1]), dtype=self.dtype)

        # set the tableau with the scratch and update
        self.tableau_w_scratch = np.vstack((tableau, scratch_row))

    @property
    def nqubits(self) -> int:
        """
        The number of qubits in the tableau.
        """
        return self.__nqubits

    @nqubits.setter
    def nqubits(self, n_qubits: int):
        assert n_qubits > 0 and isinstance(
            n_qubits, int
        ), "Number of qubits should be positive integer."

        self.__nqubits = n_qubits

    # pylint: disable=attribute-defined-outside-init
    def _define_tableau_views(self) -> None:
        """
        Defines different views of the tableau array.
        To be called upon initialization or upon adding or deleting a qubit.
        """

        self.tableau = self.tableau_w_scratch[:-1, :]
        self.scratch_row = self.tableau_w_scratch[-1, :]

        # different parts of the tableau
        self.x = self.tableau[:, : self.nqubits]
        self.z = self.tableau[:, self.nqubits : -1]
        self.r = self.tableau[:, -1]

        self.x_w_scratch = self.tableau_w_scratch[:, : self.nqubits]
        self.z_w_scratch = self.tableau_w_scratch[:, self.nqubits : -1]
        self.r_w_scratch = self.tableau_w_scratch[:, -1]

        self.stabilizer_array = self.tableau[self.nqubits :, :]
        self.destabilizer_array = self.tableau[: self.nqubits, :]

        self.z_stabilizers = self.z[self.nqubits :, :]
        self.x_stabilizers = self.x[self.nqubits :, :]
        self.r_stabilizers = self.r[self.nqubits :]

    def update(self):
        """Updates the internal structure after the tableau has
        been modified"""
        dim = self.tableau_w_scratch.shape[0]
        self.nqubits = (dim - 1) // 2

        self._define_tableau_views()

    @property
    def stabilizer_set(self) -> set[str]:
        """
        Generates the stabilizers in human readable format.
        """
        return set(StabArray(self.stabilizer_array).as_paulistrings)

    @property
    def stabilizer_set_sparse_format(self) -> list[dict]:
        """
        Returns the stabilizers in more human readable format.
        This is the sparse format.

        Note: the formatter function can be called
        by user directly as well. Look for it in QCDutils

        Parameters
        ----------

        Returns
        -------
        list[dict]
            A list of dicts, each dict representing one stabilizer operator.
            The keys refer to the Pauli operators, while value refer to the
            qubit indices where the Pauli operator resides.
            E.g. +ZXIIYXZII gets returned as
            {'sign':'+', 'X':(1,5), 'Z':(0,6), 'Y':(4,)}
        """
        return sparse_formatter(set(StabArray(self.stabilizer_array).as_paulistrings))

    def create_snapshot(self) -> TableauSnapshot:
        """
        Creates a TableauSnapshot object, a snapshot of the state of the
        Tableau, that contains important properties of the
        Tableau.

        The TableauSnapshot object can then be used to restore the state of the
        Tableau at the time the TableauSnapshot was created.
        """
        return TableauSnapshot(
            nqubits=deepcopy(self.nqubits),
            rand_gen=deepcopy(self.rand_gen),
            tableau_w_scratch=deepcopy(self.tableau_w_scratch),
        )

    def restore(self, engine_snapshot: TableauSnapshot) -> None:
        """
        This method allows us to restore the state of the Tableau
        with a TableauSnapshot Object. Once the Tableau has been
        restored, the views would be redefined.
        """

        self.nqubits = engine_snapshot.nqubits
        self.rand_gen = engine_snapshot.rand_gen
        self.tableau_w_scratch = engine_snapshot.tableau_w_scratch

        self._define_tableau_views()

    def rewrite_tableau(self, input_tableau: np.ndarray) -> Tableau:
        """
        This method updates the internal tableau of this Tableau by
        returning a new Tableau with the updated tableau. The tableau is
        updated as a whole. This method also checks for compatibility of the updated
        tableau.
        """

        # Check that new Tableau has same number of qubits as old tableau.
        # Raise printable warning.
        if not self.tableau.shape == input_tableau.shape:
            raise TableauSizeError(
                f"The Tableau has been updated from shape {self.tableau.shape} to"
                f" shape {input_tableau.shape}"
            )

        self._init_w_tableau(input_tableau=input_tableau)
        self.update()

        return self


def compare_stabilizer_set(engine_rep_1: Tableau, engine_rep_2: Tableau) -> bool:
    """Returns true if the stabilizer set of both Tableau(s) are
    equal when they are reduced.
    """
    # return false directly in case of different qubits
    if engine_rep_1.nqubits != engine_rep_2.nqubits:
        return False

    # find row echelon form of stabilizer arrays
    return is_stabarray_equivalent(
        StabArray(engine_rep_1.stabilizer_array),
        StabArray(engine_rep_2.stabilizer_array),
    )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Eka

from .base_applicator import BaseApplicator
from .measure_block_syndromes import measureblocksyndromes
from .measure_logical_pauli import measurelogicalpauli
from .logical_pauli import logical_pauli
from .reset_all_data_qubits import reset_all_data_qubits
from .reset_all_ancilla_qubits import reset_all_ancilla_qubits
from .conditional_logical_pauli import conditional_logical_pauli


class CodeApplicator(BaseApplicator):  # pylint: disable=too-few-public-methods
    """
    Contains the implementation logic for operations at the level of a code.
    CodeOperations are implemented at the level of this CodeApplicator. For more
    specific codes, subclasses are used.
    """

    def __init__(
        self,
        eka: Eka,
    ):
        super().__init__(eka)
        # Add the extra operations that are supported by the all codes
        self.supported_operations |= {
            "MeasureBlockSyndromes": measureblocksyndromes,
            "MeasureLogicalX": measurelogicalpauli,
            "MeasureLogicalY": measurelogicalpauli,
            "MeasureLogicalZ": measurelogicalpauli,
            "ResetAllDataQubits": reset_all_data_qubits,
            "LogicalX": logical_pauli,
            "LogicalY": logical_pauli,
            "LogicalZ": logical_pauli,
            "ResetAllAncillaQubits": reset_all_ancilla_qubits,
            "ConditionalLogicalX": conditional_logical_pauli,
            "ConditionalLogicalY": conditional_logical_pauli,
            "ConditionalLogicalZ": conditional_logical_pauli,
        }



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations
import dataclasses

from pydantic import Field, field_validator, ValidationInfo
from pydantic.dataclasses import dataclass

from .block import Block
from .lattice import Lattice
from .operations import Operation
from .utilities.validation_tools import dataclass_config


@dataclass(config=dataclass_config)
class Eka:
    # pylint: disable=line-too-long
    """
    This dataclass contains all the information necessary to execute a
    logical circuit with embedded error correction.

    Parameters
    ------
    lattice : Lattice
        Lattice on which the Eka is defined
    blocks : tuple[Block, ...]
        Inital state of the block defining the logical qubits.
    operations : tuple[:class:`loom.eka.operations.base_operation.Operation`, ...] | tuple[tuple[:class:`loom.eka.operations.base_operation.Operation`, ...], ...]
        Operations to be executed on the block. Each tuple of operations is executed
        in parallel. The operations are executed in the order they are given in the
        tuple but their circuits are combined such that they can be executed in
        parallel. If the operations are specified as a tuple of operations, they are
        executed sequentially.
    """

    lattice: Lattice
    blocks: tuple[Block, ...] = Field(default_factory=tuple, validate_default=False)
    operations: tuple[Operation, ...] | tuple[tuple[Operation, ...], ...] = Field(
        default_factory=tuple, validate_default=False
    )
    # hardware: Hardware # Remove comment after implementing Hardware

    @field_validator("blocks", mode="after")
    @classmethod
    def blocks_valid_indices(
        cls, blocks: tuple[Block, ...], info: ValidationInfo
    ) -> tuple[Block, ...]:
        """Check that blocks have valid qubit indices. Every element in the tuple has
        to be larger or equal to 0 and smaller than the size of the lattice in the
        respective dimension (if the lattice has a finite size)."""
        # Check if any data/ancilla qubit has a negative index
        for block in blocks:
            if any(any(coord < 0 for coord in qubit) for qubit in block.data_qubits):
                raise ValueError(
                    f"Block '{block.unique_label}' has negative data qubit indices."
                )
            if any(any(coord < 0 for coord in qubit) for qubit in block.ancilla_qubits):
                raise ValueError(
                    f"Block '{block.unique_label}' has negative ancilla qubit indices."
                )

        # Check if any data qubit has an index larger than the lattice size
        lattice_size = info.data["lattice"].size
        if lattice_size is not None:  # For infinite lattices, skip this check
            # To check that the last element in qubit indices (specifying the basis vector)
            # is valid, add the number of allowed values to `lattice_size`
            lattice_size = list(lattice_size)
            lattice_size.append(len(info.data["lattice"].basis_vectors))
            for block in blocks:
                if any(
                    any(coord >= lattice_size[i] for i, coord in enumerate(qubit))
                    for qubit in block.data_qubits
                ):
                    raise ValueError(
                        f"Block '{block.unique_label}' has data qubit indices which "
                        "are too large for the lattice."
                    )
                if any(
                    any(coord >= lattice_size[i] for i, coord in enumerate(qubit))
                    for qubit in block.ancilla_qubits
                ):
                    raise ValueError(
                        f"Block '{block.unique_label}' has ancilla qubit indices which "
                        "are too large for the lattice."
                    )

        return blocks

    _validate_blocks_no_overlap = field_validator("blocks", mode="after")(
        Block.blocks_no_overlap
    )

    @field_validator("blocks", mode="after")
    @classmethod
    def blocks_unique_labels(cls, blocks: tuple[Block, ...]) -> tuple[Block, ...]:
        """Check that blocks have unique labels."""
        unique_labels = set(block.unique_label for block in blocks)
        if len(unique_labels) != len(blocks):
            raise ValueError("Not all blocks have unique labels.")
        return blocks

    @field_validator("operations", mode="after")
    @classmethod
    def operations_disjoint(
        cls, operations: tuple[Operation] | tuple[tuple[Operation, ...], ...]
    ):
        """Check that operations are disjoint, i.e. no block is used in multiple
        operations. Casts the operations to a tuple of tuples if it is a tuple of
        operations."""
        # Cast into tuple of tuples if operations is a tuple of operations
        if all(isinstance(operation, Operation) for operation in operations):
            return tuple((operation,) for operation in operations)
        # Check that operations in the same timestep are disjoint
        for timestep in operations:
            used_inputs = set()
            for operation in timestep:
                op_inputs = set(getattr(operation, "_inputs"))
                if conflict := used_inputs.intersection(op_inputs):
                    raise ValueError(
                        f"Operations are not disjoint, {conflict} is subject to "
                        f"multiple operations at the same time."
                    )
                used_inputs.update(op_inputs)
        return operations

    @classmethod
    def fromdict(cls, data_dict: dict) -> Eka:
        """Instantiate an Eka object from a dictionary."""

        operation_input = data_dict.pop("operations", None)
        if operation_input is not None:
            data_dict["operations"] = tuple(
                tuple(Operation.fromdict(operation) for operation in timestep)
                for timestep in operation_input
            )
        return cls(**data_dict)

    def asdict(self) -> dict:
        """Create a dictionary representation of the Eka object."""
        # Leverage dataclasses.asdict excluding the 'operations' key
        class_dict = dataclasses.asdict(
            self, dict_factory=lambda d: {k: v for k, v in d if k != "operations"}
        )
        # Use custom asdict method to add the 'operations' key: value
        class_dict["operations"] = tuple(
            tuple(
                Operation.asdict(operation) for operation in timestep  # custom asdict
            )
            for timestep in self.operations
        )
        return class_dict



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from dataclasses import dataclass, field
from abc import abstractmethod

from .base_operation import Operation, OpType
from .controlled_operation import has_ccontrol


@dataclass
@has_ccontrol
class GateOperation(Operation):
    """
    Operations of this type perform quantum gate operations on qubits within the Engine
    during runtime.
    """

    operation_type: str = field(default=OpType.QUANTUMGATE, init=False)

    @property
    @abstractmethod
    def operating_qubit(self):
        """
        Returns a list of qubits that the gate operation acts on.
        """
        raise NotImplementedError


@dataclass
class OneQubitGateOperation(GateOperation):
    """
    A gate operation that acts on a single qubit.
    """

    target_qubit: int

    @property
    def operating_qubit(self):
        return [self.target_qubit]


@dataclass
class Identity(OneQubitGateOperation):
    """
    The Identity gate operation. It does nothing to the qubit.
    """

    name: str = field(default="Identity", init=False)


@dataclass
class Hadamard(OneQubitGateOperation):
    """
    The Hadamard gate operation. Acts on a single qubit.
    """

    name: str = field(default="Hadamard", init=False)


@dataclass
class Phase(OneQubitGateOperation):
    """
    The Phase gate operation. Acts on a single qubit.
    """

    name: str = field(default="Phase", init=False)


@dataclass
class PhaseInv(OneQubitGateOperation):
    """
    The Phase Inverse gate operation. Acts on a single qubit.
    """

    name: str = field(default="PhaseInv", init=False)


@dataclass
class X(OneQubitGateOperation):
    """
    The Pauli-X gate operation. Acts on a single qubit.
    """

    name: str = field(default="X", init=False)


@dataclass
class Z(OneQubitGateOperation):
    """
    The Pauli-Z gate operation. Acts on a single qubit.
    """

    name: str = field(default="Z", init=False)


@dataclass
class Y(OneQubitGateOperation):
    """
    The Pauli-Y gate operation. Acts on a single qubit.
    """

    name: str = field(default="Y", init=False)


@dataclass
class TwoQubitGateOperation(GateOperation):
    """
    A gate operation that acts on two qubits.
    """

    control_qubit: int
    target_qubit: int

    @property
    def operating_qubit(self):
        return [self.control_qubit, self.target_qubit]


@dataclass
class CNOT(TwoQubitGateOperation):
    """
    The CNOT gate operation. Acts on two qubits, the control and target qubits.
    """

    name: str = field(default="CNOT", init=False)


@dataclass
class CZ(TwoQubitGateOperation):
    """
    The CZ gate operation. Acts on two qubits, the control and target qubits.
    """

    name: str = field(default="CZ", init=False)


@dataclass
class CY(TwoQubitGateOperation):
    """
    The CY gate operation. Acts on two qubits, the control and target qubits.
    """

    name: str = field(default="CY", init=False)


@dataclass
class SWAP(TwoQubitGateOperation):
    """
    The SWAP gate operation. Acts on two qubits, swapping their states.
    """

    name: str = field(default="SWAP", init=False)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations
from functools import cached_property

import numpy as np

from .stabilizer import Stabilizer
from . import tanner_graphs


class ClassicalParityCheckMatrix:
    """Classical parity-check matrix for error-correcting codes."""

    def __init__(
        self,
        matrix_input: (
            np.ndarray
            | list[list[int]]
            | tuple[Stabilizer, ...]
            | tanner_graphs.ClassicalTannerGraph
        ),
    ):
        """
        The ClassicalParityCheckMatrix class stores a parity-check matrix
        faithfully describing a classical error-correcting code. This is a binary
        matrix, where each row describes a classical check and each column a
        classical bit.

        The object can be instantiated from a numpy array, a list of lists, or a tuple
        of Stabilizers. For an array-like input, the matrix is verified to be valid and
        cleaned afterwards. We adopt the convention of removing repeated rows and empty
        rows. For a Stabilizers input, we require all of them to be of the same pauli
        type, i.e. either X or Z. The support of the matrix rows is then built from the
        qubit support of each stabilizer.

        The ClassicalParityCheckMatrix object can also be casted as a list of
        Stabilizers.

        Lastly the matrix, can also be generated from a ClassicalTannerGraph, where
        the connectivity of the check nodes is translated into the rows of the
        matrix.

        Parameters
        ----------
        input : np.ndarray | list[list[int]] | tuple[Stabilizer,...] |
        ClassicalTannerGraph
            Input to instantiate the ClassicalParityCheckMatrix object.
        """

        if isinstance(matrix_input, np.ndarray):
            h_matrix = matrix_input
            self.verify_input(h_matrix)

        elif isinstance(matrix_input, list) and all(
            isinstance(row, list) for row in matrix_input
        ):
            h_matrix = np.array(matrix_input)
            self.verify_input(h_matrix)

        elif isinstance(matrix_input, tuple) and all(
            isinstance(item, Stabilizer) for item in matrix_input
        ):
            h_matrix = self.generate_matrix_from_stabilizers(matrix_input)
        elif isinstance(matrix_input, tanner_graphs.ClassicalTannerGraph):
            h_matrix = self.generate_matrix_from_graph(matrix_input)
        else:
            raise TypeError(
                "A numpy.array, list of list, a tuple of Stabilizers or a "
                "ClassicalTannerGraph must be provided."
            )

        self.matrix = h_matrix
        self.clean_matrix()
        self.n_checks, self.n_datas = self.matrix.shape

    @staticmethod
    def verify_input(h_matrix: np.ndarray) -> None:
        """
        Verify the input parity-check matrix is valid. The matrix is required to
        be binary, two-dimensional, and non-empty.

        Parameters
        ----------
        h_matrix : np.ndarray
            Parity-check matrix to be verified.
        """

        # Check for empty matrix
        if not np.any(h_matrix):
            raise ValueError("Parity-check matrix is empty.")

        # Ensure array is two-dimensional
        if len(h_matrix.shape) != 2:
            raise ValueError("Parity-check matrix must be a 2D array.")

        # Check for non-binary elements
        if not np.all(np.isin(h_matrix, [0, 1])):
            raise ValueError("Parity-check matrix contains non-binary elements.")

    def clean_matrix(self) -> None:
        """
        Clean the parity-check matrix by removing repeated and empty rows and
        columns.
        """

        # Remove repeated rows
        # Rows are manually reversed to counter enforced sorting from numpy.unique
        self.matrix = np.unique(self.matrix, axis=0)[::-1]

        # Remove empty rows
        self.matrix = self.matrix[~np.all(self.matrix == 0, axis=1)]

        # Remove empty columns
        self.matrix = self.matrix[:, ~np.all(self.matrix == 0, axis=0)]

    @staticmethod
    def generate_matrix_from_stabilizers(
        stabilizers: tuple[Stabilizer, ...],
    ) -> np.ndarray:
        """
        Generate parity-check matrix from a set of stabilizers.

        Parameters
        ----------
        stabilizers : tuple[Stabilizer,...]
            Stabilizers to generate the parity-check matrix from.

        Returns
        -------
        h_matrix : np.ndarray
            Parity-check matrix generated from the stabilizers.
        """

        # Verifuy non-empty input
        if len(stabilizers) == 0:
            raise ValueError("Input Stabilizer tuple is empty.")

        # Verify all stabilizers are of the same type
        pauli_type = set(p for stab in stabilizers for p in stab.pauli)
        if len(pauli_type) > 1:
            raise ValueError(
                "Input stabilizers must be of the same type to define a classical"
                " parity check matrix."
            )

        # Extract the number of data qubits
        all_data_qubits = {qubit for stab in stabilizers for qubit in stab.data_qubits}
        n_data = len(all_data_qubits)

        coord_to_index = {qubit: i for i, qubit in enumerate(all_data_qubits)}

        # Initialize parity-check matrix
        h_matrix = np.zeros((len(stabilizers), n_data), dtype=int)

        # Fill in the parity-check matrix
        for k, stabilizer in enumerate(stabilizers):
            for qubit in stabilizer.data_qubits:
                i = coord_to_index[qubit]
                h_matrix[k, i] = 1

        return h_matrix

    def to_stabilizers(self, pauli_type: str) -> list[Stabilizer]:
        """
        Convert the parity-check matrix to a list of Stabilizers.

        Parameters
        ----------
        pauli_type : str
            Pauli type to assign to the stabilizers, either 'X' or 'Z'.

        Returns
        -------
        stabilizers : list[Stabilizer]
            List of Stabilizers generated from the parity-check matrix.
        """

        # Check input
        if pauli_type not in ["X", "Z"]:
            raise ValueError("Pauli type must be either 'X' or 'Z'.")

        stabilizers = []
        for ind_row, row in enumerate(self.matrix):
            data_qubits = [(i, 0) for i in np.where(row == 1)[0]]
            ancilla_qubits = [(ind_row, 1)]
            stabilizers.append(
                Stabilizer(
                    pauli=pauli_type * len(data_qubits),
                    data_qubits=data_qubits,
                    ancilla_qubits=ancilla_qubits,
                )
            )

        return stabilizers

    @staticmethod
    def generate_matrix_from_graph(
        tanner_graph: tanner_graphs.ClassicalTannerGraph,
    ) -> np.ndarray:
        """
        Generate parity-check matrix from an input Tanner graph.

        Parameters
        ----------
        tanner_graph : graphs.ClassicalTannerGraph
            Tanner graph to generate the parity-check matrix from.

        Returns
        -------
        h_matrix : np.ndarray
            Parity-check matrix generated from the Tanner graph.
        """

        # Initialize parity-check matrix
        h_matrix = np.zeros(
            (len(tanner_graph.check_nodes), len(tanner_graph.data_nodes)), dtype=int
        )

        # Map data nodes to column indices
        data_to_ind = {d: i for i, d in enumerate(sorted(tanner_graph.data_nodes))}

        # Fill in with checks
        for i, check_node in enumerate(sorted(tanner_graph.check_nodes)):
            for data_node in tanner_graph.graph.neighbors(check_node):
                h_matrix[i, data_to_ind[data_node]] = 1

        return h_matrix

    def __eq__(self, other: ClassicalParityCheckMatrix) -> bool:
        """
        Check if two parity-check matrices are equal.

        Parameters
        ----------
        other : ClassicalParityCheckMatrix
            Other parity-check matrix to compare with.

        Returns
        -------
        bool
            True if the matrices are equal, False otherwise.
        """
        if not isinstance(other, ClassicalParityCheckMatrix):
            raise TypeError(
                "Comparison is only supported with another ClassicalParityCheckMatrix."
            )

        def are_attributes_not_the_same():
            return self.n_checks != other.n_checks or self.n_datas != other.n_datas

        # Ensure that matrices are equal up to row permutations
        def are_rows_not_the_same():
            return set(map(tuple, self.matrix)) != set(map(tuple, other.matrix))

        return not (are_attributes_not_the_same() or are_rows_not_the_same())


class ParityCheckMatrix:
    """Parity-check matrix for quantum error-correcting codes."""

    def __init__(
        self,
        input: (
            list[list] | np.ndarray | tanner_graphs.TannerGraph | tuple[Stabilizer, ...]
        ),
        # pylint: disable=redefined-builtin
    ):
        """
        The ParityCheckMatrix class stores a parity-check matrix faithfully describing a
        quantum error-correcting code. This is a binary matrix, where each row is
        associated with the symplectic representation of a stabilizer. Therefore, these
        matrices have twice the number of columns as the number of data qubits in the
        code, with the first half representing the X stabilizers and the second half
        representing the Z stabilizers. The number of rows is then equal to the number
        of stabilizers in the code. To represent a valid quantum code, these matrices
        must satisfy the vanishing symplectic product condition, i.e. the symplectic
        product of the matrix with itself must be zero. This corresponds to all the
        stabilizers commuting with each other.

        The object can be instantiated from a numpy array or a list of lists, a
        TannerGraph object or a tuple containing Stabilizer objects. For an
        array-like input, the matrix is verified to be valid and cleaned afterwards.
        We adopt the convention of removing repeated rows and empty rows.
        For a TannerGraph input, the parity-check matrix is generated from the
        connectivity of bipartite graph. For a Stabilizer input, each ancilla and each
        data qubit are mapped into rows and columns of the matrix, respectively.

        Parameters
        ----------
        input : np.ndarray | list[list[int]] | TannerGraph | tuple[Stabilizer, ...]
            Input to instantiate the ParityCheckMatrix object.
        """

        if isinstance(input, np.ndarray):
            h_matrix = input
            self.verify_input(h_matrix)

        elif isinstance(input, list) and all(isinstance(row, list) for row in input):
            h_matrix = np.array(input)
            self.verify_input(h_matrix)

        elif isinstance(input, tanner_graphs.TannerGraph):
            h_matrix = self.generate_matrix_from_graph(input)

        elif isinstance(input, tuple) and all(
            isinstance(item, Stabilizer) for item in input
        ):
            h_matrix = self.generate_matrix_from_stabilizers(input)

        else:
            raise TypeError(
                "A numpy.array, list of lists, tuple of Stabilizers or a TannerGraph "
                "object must be provided."
            )

        self.matrix = h_matrix
        self.clean_matrix()
        self.n_stabs, self.n_datas = self.matrix.shape[0], self.matrix.shape[1] // 2

        # Check if code is CSS
        self.is_css = self.check_if_css

    @staticmethod
    def verify_input(h_matrix: np.ndarray) -> None:
        """
        Verifies that parity-check matrix defines a valid quantum code, through
        structural checks and computing the symplectic product with itself.

        Parameters
        ----------
        h_matrix : np.ndarray
            A full parity-check matrix for a quantum code.
        """

        # Check for empty matrix
        if not np.any(h_matrix):
            raise ValueError("Parity-check matrix is empty.")

        # Ensure array is two-dimensional
        if len(h_matrix.shape) != 2:
            raise ValueError("Parity-check matrix must be a 2D array.")

        # Check input
        if not np.all(np.isin(h_matrix, [0, 1])):
            raise ValueError("Parity-check matrix contains non-binary elements.")

        # Check number of data qubits is well defined
        if len(h_matrix[0]) % 2 == 1:
            raise ValueError("Parity-check matrix contains odd number of columns.")

        # Extract the number of data qubits
        n = len(h_matrix[0]) // 2

        # Define the symplectic matrix
        symplectic_matrix = np.vstack(
            (
                np.hstack((np.zeros((n, n)), np.eye(n))),
                np.hstack((np.eye(n), np.zeros((n, n)))),
            )
        )

        # Compute symplectic product
        product = np.dot(h_matrix, np.dot(symplectic_matrix, h_matrix.T)) % 2

        # Specify validity as a boolean
        valid = not bool(product.any())

        if not valid:
            raise ValueError("Parity-check matrix does not define a quantum code.")

    def clean_matrix(self) -> None:
        """Clean the parity-check matrix by removing repeated and empty rows and
        columns."""

        # Remove repeated rows
        # Rows are manually reversed to counter enforced sorting from numpy.unique
        self.matrix = np.unique(self.matrix, axis=0)[::-1]

        # Remove empty rows
        self.matrix = self.matrix[~np.all(self.matrix == 0, axis=1)]

        # Remove empty columns by checking both sectors
        r = np.shape(self.matrix)[1] // 2

        # Identify columns where both column[i] and column[r+i] are all zeros
        zero_pairs = np.all(self.matrix[:, :r] == 0, axis=0) & np.all(
            self.matrix[:, r:] == 0, axis=0
        )

        # Concatenate the two halves using the mask
        self.matrix = np.hstack(
            (self.matrix[:, :r][:, ~zero_pairs], self.matrix[:, r:][:, ~zero_pairs])
        )

    @staticmethod
    def generate_matrix_from_graph(
        tanner_graph: tanner_graphs.TannerGraph,
    ) -> np.ndarray:
        """
        Generate parity-check matrix from an input Tanner graph.

        Parameters
        ----------
        tanner_graph : TannerGraph
            Tanner graph object to generate the parity-check matrix from.
        Returns
        -------
        h_matrix : np.ndarray
            Parity-check matrix generated from the Tanner graph as numpy array.
        """

        # Initialize parity-check matrix
        h_matrix = np.zeros(
            (
                len(tanner_graph.x_nodes) + len(tanner_graph.z_nodes),
                2 * len(tanner_graph.data_nodes),
            )
        )

        # Map data nodes to column indices - Datas are sorted by their coordinates
        data_to_ind = {d: i for i, d in enumerate(sorted(tanner_graph.data_nodes))}

        # Fill in with X checks
        for i, check_node in enumerate(tanner_graph.x_nodes):
            for data_node in tanner_graph.graph.neighbors(check_node):
                data_ind = data_to_ind[data_node]
                h_matrix[i, data_ind] = 1

        # Fill in with Z checks
        for i, check_node in enumerate(tanner_graph.z_nodes):
            for data_node in tanner_graph.graph.neighbors(check_node):
                data_ind = data_to_ind[data_node] + len(tanner_graph.data_nodes)
                h_matrix[i + len(tanner_graph.x_nodes), data_ind] = 1

        return h_matrix

    @staticmethod
    def generate_matrix_from_stabilizers(
        stabilizers: tuple[Stabilizer, ...],
    ) -> np.ndarray:
        """
        Generate parity-check matrix from a list of stabilizers. Each stabilizer is
        converted into a row of the matrix. For consistency, we sort the data qubits
        according to their coordinates and assign them a column index in the matrix.

        Parameters
        ----------
        stabilizers : tuple[Stabilizer,...]
            Stabilizers to generate the parity-check matrix from.

        Returns
        -------
        h_matrix : np.ndarray
            Parity-check matrix generated from the stabilizers as numpy array.
        """

        # Check that stabilizers commute
        for stab1 in stabilizers:
            for stab2 in stabilizers:
                if not stab1.commutes_with(stab2):
                    raise ValueError(
                        f"Input Stabilizers {stab1} and {stab2} do not commute."
                    )

        # Extract the number of data qubits and sort it
        all_data_qubits = sorted(
            {qubit for stab in stabilizers for qubit in stab.data_qubits}
        )
        n_data = len(all_data_qubits)

        # Initialize parity-check matrix
        h_matrix = np.zeros((len(stabilizers), 2 * n_data), dtype=int)

        # Fill in the parity-check matrix
        for k, stabilizer in enumerate(stabilizers):

            # Last element omitted as it corresponds to the sign
            h_matrix[k, :] = stabilizer.as_signed_pauli_op(all_data_qubits).array[:-1]

        return h_matrix

    def to_stabilizers(self) -> list[Stabilizer]:
        """
        Converts the parity-check matrix to a list of Stabilizers. The stabilizers are
        generated by scanning each row of the matrix and extracting data qubits from
        the position of the non-zero elements. The data qubits are assigned coordinates
        in the form `(index,0)`, where `index` is the column index of the data qubit in
        the symplectic representation of the matrix. The ancilla qubits are
        assigned coordinates in the form `(index,1)`, where `index` is the row index of
        the stabilizer in the matrix.

        Returns
        -------
        stabilizers : list[Stabilizer]
            List of Stabilizers generated from the parity-check matrix.
        """

        # Initialize the list of stabilizers
        stabilizers = []

        # Symplectic definition of the Pauli operators
        ref_string = {(0, 0): "", (1, 0): "X", (0, 1): "Z", (1, 1): "Y"}

        for ind_row, row in enumerate(self.matrix):
            # Extract the X and Z parts supports of the stabilizer
            x_row, z_row = row[: self.n_datas], row[self.n_datas :]
            pauli = "".join(
                [ref_string[(x, z)] for x, z in zip(x_row, z_row, strict=True)]
            )
            data_qubits = [
                (i, 0) for i in sorted(np.where((x_row == 1) | (z_row == 1))[0])
            ]

            ancilla_qubits = [(ind_row, 1)]
            stabilizers.append(
                Stabilizer(
                    pauli=pauli, data_qubits=data_qubits, ancilla_qubits=ancilla_qubits
                )
            )

        return stabilizers

    @cached_property
    def check_if_css(self) -> bool:
        """
        Check if the parity-check matrix defines a CSS code. Commutativity is checked
        beforehand by verifying that the symplectic product of the full matrix.

        Returns
        -------
        valid_css : bool
            True if the parity-check matrix defines a valid CSS code, False otherwise.
        """

        valid_css = True

        # For CSS, the support of every row should be non-vanishing only on one side of
        # the matrix, i.e. either X or Z stabilizers.
        for row in self.matrix:
            if np.any(row[: self.n_datas]) and np.any(row[self.n_datas :]):
                valid_css = False
                break

        return valid_css

    def get_components(
        self,
    ) -> tuple[ClassicalParityCheckMatrix, ClassicalParityCheckMatrix]:
        """
        Compute the X and Z components of the parity-check matrix, if possible.

        Returns
        -------
        hx_matrix : ClassicalParityCheckMatrix
            The X component of the parity-check matrix, containing only X stabilizers.
        hz_matrix : ClassicalParityCheckMatrix
            The Z component of the parity-check matrix, containing only Z stabilizers.
        """

        # Check if matrix has already been verified as non CSS
        if not self.is_css:
            raise ValueError(
                "Parity-check matrix cannot be split into hx_matrix and hz_matrix as "
                "there are stabilizers with mixed X and Z support, thus it does not "
                "define a CSS code."
            )

        # Extract the X and Z components of the parity-check matrix
        x_component = self.matrix[np.any(self.matrix[:, : self.n_datas], axis=1)][
            :, : self.n_datas
        ]
        z_component = self.matrix[np.any(self.matrix[:, self.n_datas :], axis=1)][
            :, self.n_datas :
        ]

        # Convert to ClassicalParityCheckMatrix objects
        hx_matrix = ClassicalParityCheckMatrix(x_component)
        hz_matrix = ClassicalParityCheckMatrix(z_component)

        return hx_matrix, hz_matrix

    @cached_property
    def hx_matrix(self) -> ClassicalParityCheckMatrix:
        """Extract the X component of the parity-check matrix."""
        hx_matrix, _ = self.get_components()
        return hx_matrix

    @cached_property
    def n_xstabs(self) -> int:
        """Extract the number of X stabilizers."""
        hx_matrix, _ = self.get_components()
        return hx_matrix.n_checks

    @cached_property
    def hz_matrix(self) -> ClassicalParityCheckMatrix:
        """Extract the Z component of the parity-check matrix."""
        _, hz_matrix = self.get_components()
        return hz_matrix

    @cached_property
    def n_zstabs(self) -> int:
        """Extract the number of Z stabilizers."""
        _, hz_matrix = self.get_components()
        return hz_matrix.n_checks

    def __eq__(self, other: ParityCheckMatrix) -> bool:
        """
        Check if two ParityCheckMatrix objects are equal.

        Parameters
        ----------
        other : ParityCheckMatrix
            Other parity-check matrix to compare with.

        Returns
        -------
        bool
            True if the matrices are equal, False otherwise.
        """

        if not isinstance(other, ParityCheckMatrix):
            raise TypeError(
                "Comparison is only supported with another ParityCheckMatrix."
            )

        def are_attributes_not_equal():
            return self.n_stabs != other.n_stabs or self.n_datas != other.n_datas

        #  that matrices are equal up to row permutations
        def are_rows_not_equal():
            return set(map(tuple, self.matrix)) != set(map(tuple, other.matrix))

        return not (are_attributes_not_equal() or are_rows_not_equal())



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from typing import Optional
import uuid
from copy import deepcopy
from dataclasses import dataclass


@dataclass(frozen=True)
class ClassicalRegisterSnapshot:
    """
    A snapshot of the classical register at a specific point in time.
    """

    name: str
    no_of_bits: int
    reg: list[tuple[str, int]]


class ClassicalRegister:
    """
    A register that stores classical bits that can be accessed at runtime by the Engine.

    Comments:
    Consider Big Registry Design with Smaller Classical Registers as Views. Similar
    design to Tableau where all quantum data is in a Tableau.

    Parameters
    ----------
    name: str
        The name of this classical register.
    no_of_bits: int
        The number of bits the register will be initialized with.
    bit_ids: Optional[list[str]]
        A list containing the ids of every bit in the classical register. The ids
        provided do not have to be uuid compatible but must be of equal length to the
        number of bits. If the bit ids are not provided, the bits will be initialized
        with randomized ids generated using uuid4.
    """

    def __init__(
        self, name: str, no_of_bits: int, bit_ids: Optional[list[str]] = None
    ) -> None:
        self.name = name
        _checked_bit_ids = self._check_bit_ids(no_of_bits, bit_ids)
        # Able to modify register and easily update all views/attributes in classical
        # register accordingly.
        self.reg = list(
            zip(_checked_bit_ids, [0 for _ in range(no_of_bits)], strict=True)
        )

    def __eq__(self, input_register: "ClassicalRegister") -> bool:
        """
        2 Classical Registers are equal if their reg and name attribute are equal.
        """
        if (
            isinstance(input_register, type(self))
            and input_register.reg == self.reg
            and input_register.name == self.name
        ):
            return True
        return False

    @property
    def reg(self) -> list[tuple[str, int]]:
        """
        The classical register represented as a list of tuples whose first entry is the
        ID of the bit and the second entry is the value of the bit.
        """
        self._valid_register()
        return self._reg

    @reg.setter
    def reg(self, completed_register: list[tuple[str, int]]) -> None:
        self._reg = completed_register
        self._valid_register()

    def _check_bit_ids(
        self, no_of_bits: int, bit_ids: Optional[list[str]]
    ) -> list[str]:
        """
        Check if bit IDs is provided.
        If not provided, new IDs are generated for every bit in the register.
        If provided, checks that the bit IDs are unique and exists for every bit in
        the register.
        """
        if bit_ids is None:
            # Generate random bit ids equal to the length of self.reg
            return [str(uuid.uuid4()) for i in range(no_of_bits)]
        if len(set(bit_ids)) != len(bit_ids):
            raise ValueError("The bit IDs provided for each bit must be unique")
        if len(bit_ids) != no_of_bits:
            raise ValueError(
                "The number of bit ids in the classical register must be equal to the "
                "number of bits."
            )
        return bit_ids

    def _valid_register(self) -> None:
        """
        The classical register should have bit values that are binary.
        The bit IDs should be valid.
        """
        for each_value in self.bit_reg:
            if each_value not in [0, 1]:
                raise ValueError("The register has a non-binary value.")
        self._check_bit_ids(self.no_of_bits, self.bit_ids)

    @property
    def no_of_bits(self) -> int:
        """
        The total number of bits in the current classical register.
        """
        return len(self._reg)

    @property
    def id_bit_reg(self) -> dict[str, int]:
        """
        A dictionary whose key, value pairs are the bit IDs and their respective bit
        values.
        """
        return dict(self._reg)

    @property
    def bit_ids(self) -> list[str]:
        """
        The bit IDs of all the bits in the classical register.
        """
        return list(self.id_bit_reg.keys())

    @property
    def bit_reg(self) -> list[int]:
        """
        The bit values of all the bits in the classical register.
        """
        return list(self.id_bit_reg.values())

    def create_snapshot(self) -> ClassicalRegisterSnapshot:
        """
        Creates a ClassicalRegisterSnapshot object, a snapshot of the state of the
        ClassicalRegister, that contains important properties of the
        ClassicalRegister.

        The ClassicalRegisterSnapshot object can then be used to restore the state of
        the ClassicalRegister at the time the snapshot was created.
        """
        return ClassicalRegisterSnapshot(
            name=deepcopy(self.name),
            no_of_bits=deepcopy(self.no_of_bits),
            reg=deepcopy(self.reg),
        )

    @classmethod
    def restore(cls, cr_snapshot: ClassicalRegisterSnapshot) -> "ClassicalRegister":
        """
        This method allows us to restore the state of the ClassicalRegister
        with a ClassicalRegisterSnapshot Object.
        The method returns an initialized ClassicalRegister with the Snapshot state.
        """

        creg = cls(name=cr_snapshot.name, no_of_bits=cr_snapshot.no_of_bits)

        creg.reg = cr_snapshot.reg

        return creg



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from functools import partial

from ...eka import Circuit, Channel
from ...cliffordsim import operations as cops


def convert_circuit_to_cliffordsim(
    input_circuit: Circuit, index_to_channel_map: dict[int, Channel]
) -> list[cops.Operation]:
    """
    Converts a Circuit object into a CliffordSim Circuit.

    NOTE: Qubit indices are remapped for now as the Circuit Object
    does not contain any information about qubit indices.

    NOTE: If input_circuit contains classical channels, additional operations will be
    added that create classical registers at the start and record the classical
    registers at the end of the circuit. The classical registers are created based on
    the labels of the classical channels. The first set of letters, separated by an _,
    in the label of the classical channel is the name of the classical register.

    Parameter
    ---------
    input_circuit: Circuit
    dqubits_dict: dict

    Returns
    -------
    list[cliffordsim.operations.Operation]
        A list containing CliffordSim Operations that represent the original
        `input_circuit`. This list can be passed into CliffordSim's Engine
        to be simulated.
    """
    channel_id_to_index_map = {
        channel.id: index for index, channel in index_to_channel_map.items()
    }

    # If classical channels exist, create classical register.
    # We follow the convention where the first set of letters, separated by an _, in
    # the label of the classical channel is the name of the classical register.
    creg_cbits = [
        (input.label.split("_")[0], input.label)
        for input in input_circuit.channels
        if input.is_classical()
    ]
    # Returns a dictionary whose keys are the classical register names and values are
    # lists of classical bit labels.
    reg_dict = {}
    for each_reg_name, each_cbit_label in creg_cbits:
        if each_reg_name not in reg_dict:
            reg_dict[each_reg_name] = []
        reg_dict[each_reg_name].append(each_cbit_label)

    # creg_op_instr will contain all the Operations that create the Classical Registers.
    # creg_record_op_instr will contain all the Operations that record the Classical
    # Registers at the end of the circuit.
    creg_op_instr = []
    creg_record_op_instr = []
    for each_reg_name, cbit_labels in reg_dict.items():
        creg_op_instr.append(
            cops.CreateClassicalRegister(
                reg_name=each_reg_name, no_of_bits=len(cbit_labels), bit_ids=cbit_labels
            )
        )
        creg_record_op_instr.append(
            cops.RecordClassicalRegister(reg_name=each_reg_name)
        )

    # pylint: disable=invalid-name
    SQ_GATE_MAP = {
        "identity": cops.Identity,
        "h": cops.Hadamard,
        "x": cops.X,
        "y": cops.Y,
        "z": cops.Z,
        "phase": cops.Phase,
        "phaseinv": cops.PhaseInv,
    }

    # pylint: disable=invalid-name
    TQ_GATE_MAP = {
        "cnot": cops.CNOT,
        "cy": cops.CY,
        "cz": cops.CZ,
        "cx": cops.CNOT,
        "swap": cops.SWAP,
    }

    # pylint: disable=invalid-name
    MEASUREMENT_OPS_MAP = {
        "measurement": cops.Measurement,
        "measurementbias0": partial(cops.Measurement, bias=0),
        "measurementbias1": partial(cops.Measurement, bias=1),
    } | {
        # All measure_{basis} operations
        f"measure_{basis.lower()}": partial(cops.Measurement, basis=basis)
        for basis in ["X", "Y", "Z"]
    }

    # pylint: disable=invalid-name
    RESET_OPS_MAP = {
        "reset": cops.Reset,
    } | {
        # All reset_{state} operations
        f"reset_{state}": partial(cops.Reset, state=state)
        for state in ["0", "1", "+", "-", "+i", "-i"]
    }

    # pylint: disable=invalid-name
    CLASSICALLY_CONTROLLED_GATES_MAP = {
        f"classically_controlled_{sq_gate_name}": (cops.ControlledOperation, op)
        for sq_gate_name, op in SQ_GATE_MAP.items()
    }

    op_map = {
        **TQ_GATE_MAP,
        **SQ_GATE_MAP,
        **MEASUREMENT_OPS_MAP,
        **RESET_OPS_MAP,
        **CLASSICALLY_CONTROLLED_GATES_MAP,
    }

    def operator_selector(item: Circuit) -> cops.Operation:
        if item.name not in op_map.keys():
            raise NotImplementedError(f'Invalid operation name "{item.name}"')
        try:
            # Get classical channel labels of circuit object.
            cc_labels = [
                each_channel.label
                for each_channel in item.channels
                if each_channel.is_classical()
            ]
            # Find the qubit indices of the corresponding quantum and ancilla channels.
            target_qubit_idx = [
                channel_id_to_index_map[input.id]
                for input in item.channels
                if input.is_quantum()
            ]

            if item.name in MEASUREMENT_OPS_MAP:
                # Measurement operations can only have one target qubit.
                if len(target_qubit_idx) != 1:
                    raise ValueError(
                        "Measurement operation can only have one target qubit."
                    )
                target_qubit = target_qubit_idx[0]

                # Measurements can be written onto a classical register if it exists.
                # No Classical Channel == No writing output to Classical Register
                # UUIDs generated for measurements since no Classical Channel available.
                if len(cc_labels) == 0:
                    return op_map[item.name](target_qubit=target_qubit)
                if len(cc_labels) == 1:
                    cc_label = cc_labels[0]
                    reg_name = next(
                        (
                            each_reg_name
                            for each_reg_name, cbit_labels in reg_dict.items()
                            if cc_label in cbit_labels
                        )
                    )

                    # BOTH the label of the measurement operation and the bit id of the
                    # classical bit == Label of the Classical Channel in the crd.Circuit
                    # measurement Object.
                    return op_map[item.name](
                        target_qubit=target_qubit,
                        label=cc_label,
                        reg_name=reg_name,
                        bit_id=cc_label,
                    )

                raise ValueError(
                    "Measurement operation can only have either one or no classical "
                    "channel."
                )
            if item.name in CLASSICALLY_CONTROLLED_GATES_MAP:
                # Prepare the Operation that will be controlled.
                inner_op = op_map[item.name][1](*target_qubit_idx)

                # Classical Control X requires a classical register and classical bit
                # to be specified.
                if len(cc_labels) != 1:
                    raise ValueError(
                        "Classically controlled operation must have exactly one "
                        "classical channel."
                    )
                cc_label = cc_labels[0]
                reg_name = next(
                    (
                        each_reg_name
                        for each_reg_name, cbit_labels in reg_dict.items()
                        if cc_label in cbit_labels
                    )
                )

                return op_map[item.name][0](
                    app_operation=inner_op,
                    reg_name=reg_name,
                    bit_id=cc_label,
                )
            # else
            # map the circuit operation onto a catalyst operation using only
            # quantum channels
            return op_map[item.name](*target_qubit_idx)

        except KeyError as exc:
            raise KeyError(f"Channel index {exc} not found in qubit_map.") from exc

    output_ops = (
        creg_op_instr
        + [
            operator_selector(subcirc)
            for tick in input_circuit.flatten().circuit
            for subcirc in tick
        ]
        + creg_record_op_instr
    )

    return output_ops



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from .interpreter import interpret_eka, cleanup_final_step
from .syndrome import Syndrome
from .detector import Detector
from .logical_observable import LogicalObservable
from .interpretation_step import InterpretationStep
from .utilities import Cbit



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from pydantic.dataclasses import dataclass

from ..eka import Stabilizer, Block, LogicalState
from ..eka.utilities import is_subset_of_stabarray, StabArray
from ..cliffordsim.engine import Engine
from ..cliffordsim.operations import UpdateTableau, Operation

from .utilities import get_parity_from_cbits
from .check_abstract import AbstractValidityCheck


@dataclass(frozen=True)
class CodeStabilizerCheckOutput:
    """
    Dataclass to store the output of the Code Stabilizer check.

    Parameters
    ----------
    missing_stabilizers: tuple[Stabilizer, ...]
        A tuple containing the code stabilizers that were not found in the output.
    stabilizers_with_incorrect_parity: tuple[Stabilizer, ...]
        A tuple containing the code stabilizers that were found in the output but have
        incorrect parity.
    """

    missing_stabilizers: tuple[Stabilizer, ...]
    stabilizers_with_incorrect_parity: tuple[Stabilizer, ...]

    def __len__(self) -> int:
        """
        Returns the total number of issues found in the check.
        """
        return len(self.missing_stabilizers) + len(
            self.stabilizers_with_incorrect_parity
        )

    def __str__(self) -> str:
        """String representation of the CodeStabilizerCheckOutput."""
        out = ""
        if self.missing_stabilizers:
            out += "- Missing Stabilizers:\n"
            out += "\n".join(str(stab) for stab in self.missing_stabilizers) + "\n"
        if self.stabilizers_with_incorrect_parity:
            out += "- Stabilizers with Incorrect Parity:\n"
            out += (
                "\n".join(str(stab) for stab in self.stabilizers_with_incorrect_parity)
                + "\n"
            )
        return out.rstrip("\n")


@dataclass(frozen=True)
class CodeStabilizerCheck(AbstractValidityCheck):
    """Dataclass to store the results of the Code Stabilizer check.

    Parameters
    ----------
    output: CodeStabilizerCheckOutput
        An object containing the output of the check.

    Properties
    ----------
    message: str
        A message indicating the result of the check. It will be empty if the check is
        valid, otherwise it will contain a message describing the issue.
    valid: bool
        True if the check is valid (i.e., no missing stabilizers and no stabilizers
        with incorrect parity), False otherwise.
    """

    output: CodeStabilizerCheckOutput

    @property
    def message(self) -> str:
        match len(self.output):
            case 0:
                return ""
            case _:
                return "Some code stabilizer(s) were not found in the output."


def check_code_stabilizers_output(
    base_cliffordsim_operations: tuple[Operation, ...],
    input_block: Block,
    output_block: Block,
    output_stabilizers_parity: dict[Stabilizer, tuple[str | int, ...]],
    output_stabilizers_with_any_value: list[Stabilizer],
    seed: int | None,
) -> CodeStabilizerCheck:
    """Checks whether the correct code stabilizers are found in the output.

    Parameters
    ----------
    base_cliffordsim_operations : tuple[ \
        :class:`loom.cliffordsim.operations.base_operation.Operation`, ...]
        The cliffordsim operations that represent the circuit to be checked.
    input_block : Block
        The Block object that represents the input code.
    output_block : Block
        The Block object that represents the output code.
    output_stabilizers_parity : list[tuple[Stabilizer, tuple[str | int, ...]]]
        A list of stabilizers that are expected to be in the output with a specific
        parity. The parity is represented as a tuple of strings and integers, where the
        strings are the labels of the classical bits where the result is stored at
        runtime, and the integers are the constant parity changes. The final parity is
        calculated by XORing the values of all of these bits.
    output_stabilizers_with_any_value : list[Stabilizer]
        The list of stabilizers that are expected to be in the output, but with
        any value.
    seed : int | None
        The seed for the cliffordsim engine.

    Returns
    -------
    CodeStabilizerCheck
        The result of the Code Stabilizer check.
    """
    all_zeros_state = LogicalState(
        [f"+Z{i}" for i in range(input_block.n_logical_qubits)]
    )

    all_zeros_tableau = all_zeros_state.get_tableau(input_block)
    # Test conditions CodeStabilizersAltered
    # by running from initial logical state |00...0>
    # Any other logical state can be used as well, but |00...0> is the simplest choice.
    operations = (UpdateTableau(all_zeros_tableau), *base_cliffordsim_operations)
    cliffordsim_engine = Engine(operations, input_block.n_data_qubits, seed=seed)
    cliffordsim_engine.run()
    # Retrieve the output stabilizer array
    out_stab_array_np = cliffordsim_engine.tableau_w_scratch.stabilizer_array
    out_stab_array = StabArray(out_stab_array_np)

    # Create dictionary to flip the parity of the stabilizers using the
    # output_stabilizer_updates
    # Initialize it with all parities being 0 for all stabilizers
    stab_parity = {stab: 0 for stab in output_block.stabilizers}
    # Update the parity of the stabilizers using the output_stabilizer_updates
    stab_parity.update(
        {
            stab: get_parity_from_cbits(cliffordsim_engine, cbits)
            for stab, cbits in output_stabilizers_parity.items()
        }
    )

    # Find the stabilizers that are to be found in the output with exact values (+/-)
    output_stabilizers_with_exact_values = [
        stab
        for stab in output_block.stabilizers
        if stab not in output_stabilizers_with_any_value
    ]

    # Find the stabilizers with exact values that are missing in the output
    existing_stabs_with_incorrect_parity = []
    missing_stabilizers = []
    for stab in output_stabilizers_with_exact_values:
        # Get the stabilizer as a signed Pauli operator
        # and its flipped version based on the parity
        stab_as_pauli_op = stab.as_signed_pauli_op(output_block.data_qubits)
        stab_as_pauli_op_with_flipped_parity = stab_as_pauli_op.with_flipped_sign()

        # If parity is 1, swap the stabilizer with its flipped version.
        if stab_parity[stab] == 1:
            stab_as_pauli_op, stab_as_pauli_op_with_flipped_parity = (
                stab_as_pauli_op_with_flipped_parity,
                stab_as_pauli_op,
            )

        if not is_subset_of_stabarray(stab_as_pauli_op, out_stab_array):
            # The stabilizer is not in the output
            if is_subset_of_stabarray(
                stab_as_pauli_op_with_flipped_parity, out_stab_array
            ):
                # If the stabilizer is in the output, but with incorrect parity,
                # add it to the list of stabilizers with incorrect parity
                existing_stabs_with_incorrect_parity += [stab]
            else:
                # If neither the stabilizer nor its flipped version is in the output,
                # it is missing
                missing_stabilizers += [stab]

    # Find the stabilizers with any (+/-) values that are missing in the output
    missing_stabilizers += [
        stab
        for stab in output_stabilizers_with_any_value
        if not is_subset_of_stabarray(
            stab.as_signed_pauli_op(output_block.data_qubits),
            out_stab_array,
        )
        and not is_subset_of_stabarray(
            stab.as_signed_pauli_op(output_block.data_qubits).with_flipped_sign(),
            out_stab_array,
        )
    ]

    return CodeStabilizerCheck(
        output=CodeStabilizerCheckOutput(
            missing_stabilizers=missing_stabilizers,
            stabilizers_with_incorrect_parity=existing_stabs_with_incorrect_parity,
        )
    )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from ..syndrome import Syndrome
from ..detector import Detector
from ..interpretation_step import InterpretationStep


def generate_detectors(
    interpretation_step: InterpretationStep,
    new_syndromes: tuple[Syndrome, ...],
) -> tuple[Detector, ...]:
    """
    Generates the detectors, by matching the new syndromes to their associated old
    syndromes. This relation is obtained via the stabilizer flow specified in the
    operation applicator.

    Detectors inherit the labels from the new syndromes.

    Parameters
    ----------
    interpretation_step: InterpretationStep
        The updated interpretation step implementing the operation.
    new_syndromes: tuple[Syndrome, ...]
        The newly generated syndromes during the operation.

    Returns
    -------
    tuple[Detector, ...]
        Detectors associated with the newly generated syndromes in the operation.
    """

    # Extract old syndromes
    old_syndromes = tuple(
        interpretation_step.get_prev_syndrome(syndrome.stabilizer, syndrome.block)
        for syndrome in new_syndromes
    )

    # Get new detectors
    new_detectors = tuple(
        Detector(
            syndromes=old_syndromes_list + [new_syndrome], labels=new_syndrome.labels
        )
        for old_syndromes_list, new_syndrome in zip(
            old_syndromes, new_syndromes, strict=True
        )
        if old_syndromes_list
    )

    return new_detectors



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=duplicate-code
from __future__ import annotations
from uuid import uuid4
from functools import cached_property
from pydantic.dataclasses import dataclass

from loom.eka import (
    Block,
    Lattice,
    LatticeType,
    PauliOperator,
    Stabilizer,
    SyndromeCircuit,
    Circuit,
    Channel,
)
from loom.eka.utilities import dataclass_config


@dataclass(config=dataclass_config)
class ShorCode(Block):
    """
    A sub-class of ``Block`` that represents a Shor code block.
    """

    @classmethod
    def create(
        cls,
        lattice: Lattice,
        unique_label: str | None = None,
        position: tuple[int, ...] = (0,),
        skip_validation: bool = False,
    ) -> ShorCode:
        """Create datablock of Shor Code as a Block object. The Shor Code is a
        [[9, 1, 3]] quantum error-correcting CSS code, resulting from the concatenation
        of two repetition codes. There are two X stabilizers of weight 6, and six Z
        stabilizers of weight 2. We implement a naive syndrome extraction scheme, where
        measure all X stabilizers first and all the Z stabilizers later. We choose to
        embed the code in a linear lattice, where the logical operators are defined on
        all the data qubits.

        Parameters
        ----------
        lattice : Lattice
            Lattice on which the block is defined. The qubit indices depend on the type
            of lattice.
        unique_label : str, optional
            Label for the block. It must be unique among all blocks in the initial CRD.
            If no label is provided, a unique label is generated automatically using the
            uuid module.
        position : tuple[int], optional
            Position of the top left boundary of the block on the lattice,
            by default (0,).
        skip_validation : bool, optional
            Skip validation of the block object, by default False.

        Returns
        -------
        Block
            Block object for the Shor Code.
        """

        # Input validation
        if lattice.lattice_type != LatticeType.LINEAR:
            raise ValueError(
                "The creation of Shor code blocks is "
                "currently only supported for 1D linear lattices. Instead "
                f"the lattice is of type {lattice.lattice_type}."
            )

        if not isinstance(position, tuple) or any(
            not isinstance(x, int) for x in position
        ):
            raise ValueError(
                f"`position` must be a tuple of integers. Got '{position}' instead."
            )

        if unique_label is None:
            unique_label = str(uuid4())

        # Define the stabilizers
        x_indices = [[(i + 3 * j, 0) for i in range(6)] for j in range(2)]
        x_stabilizers = [
            Stabilizer(
                pauli="X" * 6,
                data_qubits=x_ind,
                ancilla_qubits=[(i, 1)],
            )
            for i, x_ind in enumerate(x_indices)
        ]

        z_indices = [
            [(i + j + 3 * k, 0) for i in range(2)] for k in range(3) for j in range(2)
        ]
        z_stabilizers = [
            Stabilizer(
                pauli="Z" * 2,
                data_qubits=z_ind,
                ancilla_qubits=[(i + 2, 1)],
            )
            for i, z_ind in enumerate(z_indices)
        ]

        stabilizers = x_stabilizers + z_stabilizers

        # Define the syndrome circuits
        x_syndrome_circuits = cls.generate_syndrome_extraction_circuits("X" * 6)
        z_syndrome_circuits = cls.generate_syndrome_extraction_circuits("Z" * 2)
        syndrome_circuits = [x_syndrome_circuits, z_syndrome_circuits]

        # Define mapping of stabilizers to syndrome circuits
        stabilizer_to_circuit = {
            stab.uuid: x_syndrome_circuits.uuid for stab in x_stabilizers
        } | {stab.uuid: z_syndrome_circuits.uuid for stab in z_stabilizers}

        # Define the logical operators
        all_datas = [(i, 0) for i in range(9)]
        logical_x_operator = PauliOperator(pauli="X" * 9, data_qubits=all_datas)
        logical_z_operator = PauliOperator(pauli="Z" * 9, data_qubits=all_datas)

        block = cls(
            unique_label=unique_label,
            stabilizers=stabilizers,
            logical_x_operators=[logical_x_operator],
            logical_z_operators=[logical_z_operator],
            syndrome_circuits=syndrome_circuits,
            stabilizer_to_circuit=stabilizer_to_circuit,
            skip_validation=skip_validation,
        )

        if position == (0,):
            return block

        return block.shift(position)

    # Syndrome extraction circuits
    @staticmethod
    def generate_syndrome_extraction_circuits(pauli: str) -> SyndromeCircuit:
        """
        Construct a syndrome extraction circuit for the given Pauli operator. We chose
        the convention of measuring all X stabilizers first and then all Z, thus,
        appropriate idling steps are added depending on the type of Pauli operator.

        Parameters
        ----------
        pauli : str
            The Pauli operator as a string.

        Returns
        -------
        syndrome_circuit : SyndromeCircuit
            The syndrome extraction circuit for the given Pauli operator.
        """

        # Extract pauli information
        pauli_type = pauli[0]
        supp = len(pauli)
        circ_name = f"{pauli}_syndrome_extraction"

        # Define channels
        data_channels = [Channel(type="quantum", label=f"d{i}") for i in range(supp)]
        cbit_channel = Channel(type="classical", label="c0")
        ancilla_channel = Channel(type="quantum", label="a0")

        # Add ancilla reset
        reset = [Circuit("Reset_0", channels=[ancilla_channel])]

        # Hadamard layers
        h_initial = [Circuit("H", channels=[ancilla_channel])]
        h_final = [Circuit("H", channels=[ancilla_channel])]

        # Add entangling layer
        entangling_layer = [
            [Circuit(f"C{p}", channels=[ancilla_channel, data_channels[i]])]
            for i, p in enumerate(pauli)
        ]

        ## Add idling steps
        if pauli_type == "X":
            entangling_layer += [()] * 2
        else:
            entangling_layer = [()] * 6 + entangling_layer

        # Add ancilla measurement
        measurement = [Circuit("Measurement", channels=[ancilla_channel, cbit_channel])]

        circuit_list = [reset, h_initial] + entangling_layer + [h_final, measurement]

        syndrome_circuit = SyndromeCircuit(
            pauli=pauli,
            name=circ_name,
            circuit=Circuit(
                name=circ_name,
                circuit=circuit_list,
                channels=data_channels + [ancilla_channel, cbit_channel],
            ),
        )

        return syndrome_circuit

    # Instance methods
    def __eq__(self, other: ShorCode) -> bool:
        if not isinstance(other, ShorCode):
            raise NotImplementedError(f"Cannot compare ShorCode with {type(other)}")
        return super().__eq__(other)

    def shift(
        self, position: tuple[int, ...], new_label: str | None = None
    ) -> ShorCode:
        return super().shift(position, new_label)

    def rename(self, name: str) -> ShorCode:
        return super().rename(name)

    @cached_property
    def stabilizers_labels(self) -> dict[str, dict[str, tuple[int, ...]]]:
        return super().stabilizers_labels



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from ..eka import Circuit, Channel, Block, Stabilizer, LogicalState
from ..executor.legacy_functions import convert_circuit_to_cliffordsim
from ..cliffordsim.operations import DeleteQubit, AddQubit, SWAP, Operation

from .debug_dataclass import DebugData, AllChecks

from .check_stabilizer_measurement import check_input_stabilizer_measurement
from .check_logical_ops import check_logical_operators_transformation
from .check_code_stabilizers import check_code_stabilizers_output


# pylint: disable=too-many-arguments, too-many-positional-arguments
def is_circuit_valid(
    circuit: Circuit,
    input_block: Block | tuple[Block, ...],
    output_block: Block | tuple[Block, ...],
    output_stabilizers_parity: dict[Stabilizer, tuple[str | int, ...]],
    output_stabilizers_with_any_value: list[Stabilizer],
    logical_state_transformations_with_parity: dict[
        LogicalState,
        tuple[LogicalState, dict[int, tuple[str | int, ...]]],
    ],
    logical_state_transformations: list[tuple[LogicalState, tuple[LogicalState, ...]]],
    measurement_to_input_stabilizer_map: dict[str, Stabilizer],
    seed: int | None = None,
) -> DebugData:
    """
    Tests if a QEC circuit is valid, ie if it meets the criteria:
    
    - The input code is transformed into the output code
    
        - If any generators are allowed to have any value, positive or negative, the \
        ``output_stabilizers_with_any_value`` list should contain them
        - If the generators are not allowed to have any value, but their parity is \
        specified, the ``output_stabilizers_parity`` list should contain the classical \
        channels and integer constant flips(0 or 1) that are applied to them.

    - The ``logical_state_transformations_with_parity`` dictionary should contain:
    
        - The input logical state as the key
        - A tuple as the value where:
        
            - The first element is the output logical state
            - The second is a dictionary where:
            
                - The keys are the logical operator indices
                - The values are lists of strings or integers (0 or 1) that represent \
                the classical channels and parity flips applied to the logical \
                operators.
                
    - The input code stabilizers are measured correctly

    Parameters
    ----------
    circuit : Circuit
        The QEC circuit.
    input_block : Block | tuple[Block, ...]
        The input Block object(s) corresponding to the input code.
    output_block : Block | tuple[Block, ...]
        The output Block object(s) corresponding to the output code.
    output_stabilizers_parity : dict[Stabilizer, tuple[str | int, ...]]
        Dictionary where the keys are output stabilizers and the value is the expected
        parity. The parity is represented as a tuple of strings and integers (0 or 1),
        where the strings are the labels of the classical bits where a result is stored
        at runtime, and the integers are the constant parity changes. The final parity
        is calculated by XORing the values of all of these bits.
    output_stabilizers_with_any_value : list[Stabilizer]
        List of output code generators that are allowed to have any value in the end of
        the circuit.
    logical_state_transformations_with_parity : dict[
        LogicalState,
        tuple[LogicalState, dict[int, tuple[str | int, ...]]],
    ]
        Dictionary where the keys are the input logical states and the values are
        tuples containing the output logical state and a dictionary of parity flips
        that correspond to each logical operator. The keys of the dictionary are the
        logical operator indices, and the values are lists of strings or integers (0 or
        1) that represent the classical channels and parity flips applied to the logical
        operators.
    logical_state_transformations :
    list[tuple[LogicalState, tuple[LogicalState, ...]]]
        List of tuples where each tuple contains the transformation of the logical
        operators from the input state to the output state(s) that will be checked.
    measurement_to_input_stabilizer_map : dict[str, Stabilizer]
        Dictionary matching the classical channel name of a measurement operation with a
        stabilizer in the input code.
    seed : int, optional
        The seed for the cliffordsim engine, by default None. None means that the
        cliffordsim engine will not be seeded and will use a random seed.

    Returns
    -------
    DebugData
        The result of the checks.
    """
    # If input or output block is a tuple, convert it to a single Block object
    if not isinstance(input_block, Block):
        input_block = Block.from_blocks(input_block)
    if not isinstance(output_block, Block):
        output_block = Block.from_blocks(output_block)

    ### VALIDATION OF INPUTS ###

    # check that  measurement_to_input_stabilizer_map indeed has Stabilizer as values
    validate_measurement_to_input_stabilizer_map(
        circuit, input_block, measurement_to_input_stabilizer_map
    )
    # check that output_stabilizers_with_any_value indeed has Stabilizer objects that
    # are part of the output block(s)
    validate_output_stabilizers_with_any_value(
        output_block, output_stabilizers_with_any_value
    )
    # Check that logical_state_transformations_with_parity has the correct type
    validate_logical_state_transformations_with_parity(
        circuit,
        input_block,
        output_block,
        logical_state_transformations_with_parity,
    )
    # Check Logical State Transformations input
    validate_logical_state_transformations(
        input_block, output_block, logical_state_transformations
    )
    # Check that output_stabilizers_parity has the correct type
    validate_output_stabilizers_parity(circuit, output_block, output_stabilizers_parity)

    ### CLIFFORDSIM OPERATIONS ###

    # Obtain the cliffordsim operations for the circuit to run validator
    base_cliffordsim_operations = get_validator_cliffordsim_operations(
        circuit, input_block, output_block
    )

    ### RUN THE VALIDATOR CHECKS ###

    # Test if the code stabilizers were correctly transformed.
    code_stab_check = check_code_stabilizers_output(
        base_cliffordsim_operations,
        input_block,
        output_block,
        output_stabilizers_parity,
        output_stabilizers_with_any_value,
        seed,
    )
    # Test if the logical operators were altered
    log_op_check = check_logical_operators_transformation(
        base_cliffordsim_operations,
        input_block,
        output_block,
        logical_state_transformations_with_parity,
        logical_state_transformations,
        seed,
    )
    # Test if the syndrome was correctly measured
    stab_meas_check = check_input_stabilizer_measurement(
        base_cliffordsim_operations,
        input_block,
        measurement_to_input_stabilizer_map,
        seed,
    )

    return DebugData(AllChecks(code_stab_check, log_op_check, stab_meas_check))


### VALIDATION FUNCTIONS ###


def validate_measurement_to_input_stabilizer_map(
    circuit: Circuit,
    input_block: Block,
    measurement_to_input_stabilizer_map: dict[str, Stabilizer],
) -> None:
    """
    Validate the measurement_to_input_stabilizer_map input and raise an appropriate
    error if invalid.

    Parameters
    ----------
    circuit : Circuit
        The QEC circuit.
    input_block : Block
        The input Block object corresponding to the input code.
    measurement_to_input_stabilizer_map : dict[str, Stabilizer]
        Dictionary matching the classical channel name of a measurement operation with a
        stabilizer in the input code.
    """

    if not isinstance(measurement_to_input_stabilizer_map, dict):
        raise TypeError(
            "The measurement_to_input_stabilizer_map should be of type dict."
            f"Value {measurement_to_input_stabilizer_map} is of type "
            f"{type(measurement_to_input_stabilizer_map)}."
        )

    classical_channel_labels = [
        chan.label for chan in circuit.channels if chan.is_classical()
    ]

    for cchan_label, stab in measurement_to_input_stabilizer_map.items():
        # check that measurement_to_input_stabilizer_map indeed has Stabilizer as values
        if not isinstance(stab, Stabilizer):
            raise TypeError(
                "The values of measurement_to_input_stabilizer_map should be of type "
                f"Stabilizer. Value {stab} is of type {type(stab)}."
            )
        if stab not in input_block.stabilizers:
            raise ValueError(f"Stabilizer {stab} was not found in the input block.)")
        if not isinstance(cchan_label, str):
            raise TypeError(
                "The keys of measurement_to_input_stabilizer_map should be of type "
                f"str. Key {cchan_label} is of type {type(cchan_label)}."
            )
        if cchan_label not in classical_channel_labels:
            raise ValueError(
                f"Classical channel label {cchan_label} was not found in the circuit's "
                "classical channels."
            )


def validate_output_stabilizers_with_any_value(
    output_block: Block,
    output_stabilizers_with_any_value: list[Stabilizer],
) -> None:
    """
    Validate the output_stabilizers_with_any_value input and raise an appropriate
    error if invalid.

    Parameters
    ----------
    output_block : Block
        The output Block object corresponding to the output code.
    output_stabilizers_with_any_value : list[Stabilizer]
        List of output code generators that are allowed to have any value in the end of
        the circuit.
    """
    if any(
        stab not in output_block.stabilizers
        for stab in output_stabilizers_with_any_value
    ):
        raise ValueError(
            "The stabilizers in output_stabilizers_with_any_value should be in the "
            "stabilizers of the output block(s)."
        )


def validate_output_stabilizers_parity(
    circuit: Circuit,
    output_block: Block,
    output_stabilizers_parity: dict[Stabilizer, tuple[str | int, ...]],
) -> None:
    """Validate the output_stabilizers_parity input and raise an appropriate error if
    invalid.

    Parameters
    ----------
    circuit : Circuit
        The QEC circuit.
    output_block : Block
        The output Block object corresponding to the output code.
    output_stabilizers_parity : dict[Stabilizer, tuple[str | int, ...]]
        Dictionary where the keys are output stabilizers and the value is the expected
        parity. The parity is represented as a tuple of strings and integers (0 or 1),
        where the strings are the labels of the classical bits where a result is stored
        at runtime, and the integers are the constant parity changes. The final parity
        is calculated by XORing the values of all of these bits.
    """
    if not isinstance(output_stabilizers_parity, dict):
        raise TypeError("The output_stabilizers_parity should be a dictionary.")

    for key_stab, value_parity in output_stabilizers_parity.items():
        if not isinstance(key_stab, Stabilizer):
            raise TypeError(
                "The keys of output_stabilizers_parity should be of type Stabilizer."
                f" Key {key_stab} is of type {type(key_stab)}."
            )

        if not key_stab in output_block.stabilizers:
            raise ValueError(
                "Every stabilizer in output_stabilizers_parity should be in the "
                f"stabilizers of the output block. {key_stab} was not found in the"
                f" output block."
            )
        if not isinstance(value_parity, (tuple, list)):
            raise TypeError(
                "The values of the output_stabilizers_parity should be a tuple or "
                f"a list. Value {value_parity} is of type {type(value_parity)}."
            )

        for cbit in value_parity:
            if not isinstance(cbit, (str, int)):
                raise TypeError(
                    "Every element of the output_stabilizers_parity value should be "
                    "either a string (classical channel label) or an integer (0 or 1). "
                    f"Element {cbit} is of type {type(cbit)}."
                )
            if isinstance(cbit, str):
                is_channel_in_output = any(
                    chan.label == cbit and chan.is_classical()
                    for chan in circuit.channels
                )
                if not is_channel_in_output:
                    raise ValueError(
                        f"Channel {cbit} which is referenced in the "
                        f"output_stabilizers_parity is not a classical channel in the "
                        f"circuit."
                    )
            if isinstance(cbit, int):
                if cbit not in (0, 1):
                    raise ValueError(
                        f"Only 0 and 1 are allowed as parity changes. {cbit} was found."
                    )


# pylint: disable=too-many-branches
def validate_logical_state_transformations_with_parity(
    circuit: Circuit,
    input_block: Block,
    output_block: Block,
    logical_state_transformations_with_parity: dict[
        LogicalState,
        tuple[LogicalState, dict[int, tuple[str | int, ...] | list[str | int]]],
    ],
) -> None:
    """
    Validate the logical_state_transformations_with_parity input and raise an
    appropriate error if invalid.

    Parameters
    ----------
    circuit : Circuit
        The QEC circuit.
    input_block : Block
        The input Block object corresponding to the input code.
    output_block : Block
        The output Block object corresponding to the output code.
    logical_state_transformations_with_parity : dict[
        LogicalState,
        tuple[LogicalState, dict[int, tuple[str | int, ...] | list[str | int]]],
    ]
        Dictionary where the keys are the input logical states and the values are
        tuples containing the output logical state and a dictionary of parity flips
        that correspond to each logical operator. The keys of the dictionary are the
        logical operator indices, and the values are lists of strings or integers (0 or
        1) that represent the classical channels and parity flips applied to the logical
        operators.
    """

    for logical_state_in, (
        logical_state_out,
        parity_flips,
    ) in logical_state_transformations_with_parity.items():

        ## Check Key
        if not isinstance(logical_state_in, LogicalState):
            raise TypeError(
                "The input of logical_state_transformations_with_parity should be of "
                f"type LogicalState. Element {logical_state_in} is of type "
                f"{type(logical_state_in)}."
            )
        if not isinstance(logical_state_out, LogicalState):
            raise TypeError(
                "The output of logical_state_transformations_with_parity should be of "
                f"type LogicalState. Element {logical_state_out} is of type "
                f"{type(logical_state_out)}."
            )

        # Check dimensional compatibility
        if logical_state_in.n_logical_qubits != input_block.n_logical_qubits:
            raise ValueError(
                f"Input code has {input_block.n_logical_qubits} logical qubits, "
                f"but a logical state has {logical_state_in.n_logical_qubits}."
            )
        if logical_state_out.n_logical_qubits != output_block.n_logical_qubits:
            raise ValueError(
                f"Output code has {output_block.n_logical_qubits} logical qubits, "
                f"but a logical state has {logical_state_out.n_logical_qubits}."
            )

        ## Check Value
        if not isinstance(parity_flips, dict):
            raise TypeError(
                "The second element of each tuple in "
                "logical_state_transformations_with_parity should be a dictionary."
            )
        if not all(isinstance(key, int) for key in parity_flips.keys()):
            raise TypeError(
                "The keys of the second element of each tuple in "
                "logical_state_transformations_with_parity should be of type int."
            )

        for logical_idx, cbits_list in parity_flips.items():
            if logical_idx < 0 or logical_idx >= logical_state_out.n_logical_qubits:
                raise ValueError(
                    "The keys of the second element of each tuple in "
                    "logical_state_transformations_with_parity should be in the range "
                    f"of 0 to {logical_state_out.n_logical_qubits - 1}."
                    f" Found {logical_idx}."
                )
            if not isinstance(cbits_list, (tuple, list)):
                raise TypeError(
                    "The second element of each tuple in "
                    "logical_state_transformations_with_parity should be a list or "
                    f"tuple. Element {logical_idx} is of type {type(cbits_list)}."
                )
            for cbit in cbits_list:
                if not isinstance(cbit, (str, int)):
                    raise TypeError(
                        "The value of the second element of each tuple in "
                        "logical_state_transformations_with_parity should be an "
                        "iterable of strings or ints. "
                        f"Found {cbit} of type {type(cbit)}."
                    )
                if isinstance(cbit, str):
                    is_channel_in_output = any(
                        chan.label == cbit and chan.is_classical()
                        for chan in circuit.channels
                    )
                    if not is_channel_in_output:
                        raise ValueError(
                            f"Channel {cbit} which is referenced in the "
                            "logical_state_transformations_with_parity is not a "
                            "classical channel in the output block."
                        )
                if isinstance(cbit, int):
                    if cbit not in (0, 1):
                        raise ValueError(
                            f"Only 0 and 1 are allowed as parity changes in"
                            f"logical_state_transformations_with_parity. {cbit} "
                            "was found."
                        )


def validate_logical_state_transformations(
    input_block: Block,
    output_block: Block,
    logical_state_transformations: list[tuple[LogicalState, tuple[LogicalState, ...]]],
) -> None:
    """
    Validate the logical_state_transformations input and raise an appropriate error
    if invalid.

    Parameters
    ----------
    input_block : Block
        The input Block object corresponding to the input code.
    output_block : Block
        The output Block object corresponding to the output code.
    logical_state_transformations : list[tuple[LogicalState, tuple[LogicalState, ...]]]
        List of tuples where each tuple contains the transformation of the logical
        operators from the input state to the output state(s) that will be checked.
    """
    for logical_state_in, logical_states_out in logical_state_transformations:
        # Check Type
        if not isinstance(logical_state_in, LogicalState):
            raise TypeError(
                "The input of logical_state_transformations should be of type "
                "LogicalState."
            )

        if not all(
            isinstance(logical_state, LogicalState)
            for logical_state in logical_states_out
        ):
            raise TypeError(
                "The output of logical_state_transformations should be a tuple of "
                "LogicalState objects."
            )

        # Check dimensional compatibility
        if logical_state_in.n_logical_qubits != input_block.n_logical_qubits:
            raise ValueError(
                f"Input code has {input_block.n_logical_qubits} logical qubits, "
                f"but a logical state has {logical_state_in.n_logical_qubits}."
            )
        for logical_state in logical_states_out:
            if logical_state.n_logical_qubits != output_block.n_logical_qubits:
                raise ValueError(
                    f"Output code has {output_block.n_logical_qubits} logical "
                    f"qubits, but a logical state has {logical_state.n_logical_qubits}."
                )


### CLIFFORDSIM INSTRUCTIONS UTILITY FUNCTIONS ###


def find_final_swaps(
    input_all_qubit_to_channel_map: dict[int, Channel],
    output_data_qubit_to_channel_map: dict[int, Channel],
) -> list[SWAP]:
    """Find the SWAP operations to bring the final qubits into the correct position.

    Parameters
    ----------
    input_all_qubit_to_channel_map : dict[int, Channel]
        A dictionary matching all qubit indices with their Channel at the beginning
        of the circuit
    output_data_qubit_to_channel_map : dict[int, Channel]
        A dictionary matching data qubit indices with their Channel at the end of
        the circuit.

    Returns
    -------
    list[SWAP]
        A list of SWAP operations to bring the final qubits into the correct position.
    """
    # find channels that don't have the correct data qubit already
    channels_with_wrong_dq = {
        chan: idx
        for idx, chan in output_data_qubit_to_channel_map.items()
        if input_all_qubit_to_channel_map[idx] != chan
    }

    # initialize the map with the data qubits
    channel_to_idx_map = {
        chan: idx for idx, chan in input_all_qubit_to_channel_map.items()
    }

    # make an inverse map for channel to index
    idx_to_chan_map = {idx: chan for chan, idx in channel_to_idx_map.items()}

    # SWAP the final qubits into the correct position
    swap_operations = []
    while channels_with_wrong_dq:
        # pop one by one the items from the dictionary
        chan, dqubit_idx = channels_with_wrong_dq.popitem()

        chan_qubit_idx = channel_to_idx_map[chan]
        dqubit_chan = idx_to_chan_map[dqubit_idx]

        if chan_qubit_idx != dqubit_idx:
            # if the qubits are not in the correct position, swap them
            swap_operations.append(SWAP(dqubit_idx, chan_qubit_idx))

            # update maps accordingly (both ways)
            channel_to_idx_map[chan] = dqubit_idx
            channel_to_idx_map[dqubit_chan] = chan_qubit_idx

            idx_to_chan_map[dqubit_idx] = chan
            idx_to_chan_map[chan_qubit_idx] = dqubit_chan

    return swap_operations


# pylint: disable=anomalous-backslash-in-string
def get_validator_cliffordsim_operations(
    circuit: Circuit,
    input_block: Block,
    output_block: Block,
) -> tuple[Operation, ...]:
    """Generate a list of cliffordsim operations from a circuit and some input and
    output data qubit to channel maps.
    The qubit channels of the circuit that are not in the input map are considered to be
    auxiliary qubits and initialized in state :math:`\ket{0}`.
    The qubit channels of the output are swapped into the correct position as dictated
    by the output map.
    The qubit channels that are not in the output map are deleted at the end of the
    circuit.

    Parameters
    ----------
    circuit : Circuit
        The base circuit to be converted to cliffordsim instructions.
    input_block : Block
        The Block object containing the input stabilizers and logical operators.
    output_block : Block
        The Block object containing the output stabilizers and logical operators.

    Returns
    -------
    tuple[:class:`loom.cliffordsim.operations.base_operation.Operation`, ...]
        A tuple of cliffordsim operations to be executed.
    """

    # Get all qubit channels
    circuit_qubit_channels = [inp for inp in circuit.channels if inp.is_quantum()]

    # Find the data qubit labels
    input_block_data_qubits_str = [str(qubit) for qubit in input_block.data_qubits]
    # Find all qubit channel labels
    circuit_qubit_channels_labels = [chan.label for chan in circuit_qubit_channels]

    # Find which input data qubits are not indexed
    input_qubits_not_indexed = [
        qub
        for qub in input_block_data_qubits_str
        if qub not in circuit_qubit_channels_labels
    ]

    n_qubits = len(circuit_qubit_channels) + len(input_qubits_not_indexed)
    # Count data qubits and auxiliary qubits for input and output
    n_dqubits_in = input_block.n_data_qubits
    n_auxqubits_in = n_qubits - n_dqubits_in
    n_dqubits_out = output_block.n_data_qubits
    n_auxqubits_out = n_qubits - n_dqubits_out

    # Define the input data qubit channel map
    input_data_qubit_to_channel_map = {
        input_block_data_qubits_str.index(chan.label): chan
        for chan in circuit_qubit_channels
        if chan.label in input_block_data_qubits_str
    }

    # Define ancilla qubit channel map
    input_aux_qubit_channels = [
        chan
        for chan in circuit_qubit_channels
        if chan.label not in input_block_data_qubits_str
    ]
    input_aux_qubit_to_channel_map = {
        idx + n_dqubits_in: chan for idx, chan in enumerate(input_aux_qubit_channels)
    }

    # Get all qubit to channel map
    input_all_qubit_to_channel_map = {
        **input_data_qubit_to_channel_map,
        **input_aux_qubit_to_channel_map,
        # Add qubit channels that are not indexed
        **{
            input_block_data_qubits_str.index(qubit): Channel(label=qubit)
            for qubit in input_qubits_not_indexed
        },
    }

    # Define the output data qubit channel map
    output_data_qubits_str = [str(qubit) for qubit in output_block.data_qubits]
    output_data_qubit_to_channel_map = {
        output_data_qubits_str.index(chan.label): chan
        for chan in list(input_all_qubit_to_channel_map.values())
        if chan.label in output_data_qubits_str
    }

    ### CONSTRUCT THE CLIFFORDSIM INSTRUCTION LIST ###

    # Start with the initialization of auxiliary qubits
    cliffordsim_instructions = [AddQubit(n_dqubits_in) for _ in range(n_auxqubits_in)]

    # Append the circuit instructions
    cliffordsim_instructions += convert_circuit_to_cliffordsim(
        circuit, input_all_qubit_to_channel_map
    )

    # Add the SWAP operations to bring the final qubits into the correct position
    cliffordsim_instructions += find_final_swaps(
        input_all_qubit_to_channel_map,
        output_data_qubit_to_channel_map,
    )

    # Append the deletion of auxiliary qubits
    cliffordsim_instructions += [
        DeleteQubit(n_dqubits_out) for _ in range(n_auxqubits_out)
    ]

    return tuple(cliffordsim_instructions)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from pydantic.dataclasses import dataclass
from .base_operation import Operation
from ..utilities import SingleQubitPauliEigenstate, dataclass_config


@dataclass(config=dataclass_config)
class LogicalOperation(Operation):
    """
    Parent class for all logical operations in the Eka. All logical operations acts
    on logical qubits. Details of implementation are abstracted away, including how the
    qubits are constructed. E.g. they may be constructed from multiple blocks with a
    single qubit each, or from a single block with multiple qubits.

    Note that these operations only modify the state of the logical qubits, they do not
    perform any code level operations like making sure the structure of the code is
    preserved (e.g. a block ends up being rotated).
    """


@dataclass(config=dataclass_config)
class Reset(LogicalOperation):
    """
    Reset a logical qubit to one of the supported states.

    Parameters
    ----------
    target_qubit : str
        Name of the logical qubit to reset.
    state : SingleQubitPauliEigenstate
        State to reset the qubit to.
    """

    target_qubit: str
    state: SingleQubitPauliEigenstate


@dataclass(config=dataclass_config)
class CNOT(LogicalOperation):
    """
    Describes a CNOT gate between two logical qubits.

    Parameters
    ----------
    control_qubit : str
        Name of the control logical qubit.
    target_qubit : str
        Name of the target logical qubit.
    """

    control_qubit: str
    target_qubit: str


@dataclass(config=dataclass_config)
class Hadamard(LogicalOperation):
    """
    Describes a Hadamard gate on a logical qubit.

    Parameters
    ----------
    target_qubit : str
        Name of the target logical qubit.
    """

    target_qubit: str


@dataclass(config=dataclass_config)
class Phase(LogicalOperation):
    """
    Describes a Phase gate on a logical qubit.

    Parameters
    ----------
    target_qubit : str
        Name of the target logical qubit.
    """

    target_qubit: str


@dataclass(config=dataclass_config)
class PhaseInverse(LogicalOperation):
    """
    Describes an inverse Phase gate on a logical qubit.

    Parameters
    ----------
    target_qubit : str
        Name of the target logical qubit.
    """

    target_qubit: str


@dataclass(config=dataclass_config)
class X(LogicalOperation):
    """
    Describes an X gate on a logical qubit.

    Parameters
    ----------
    target_qubit : str
        Name of the target logical qubit.
    """

    target_qubit: str


@dataclass(config=dataclass_config)
class Y(LogicalOperation):
    """
    Describes a Y gate on a logical qubit.

    Parameters
    ----------
    target_qubit : str
        Name of the target logical qubit.
    """

    target_qubit: str


@dataclass(config=dataclass_config)
class Z(LogicalOperation):
    """
    Describes a Z gate on a logical qubit.

    Parameters
    ----------
    target_qubit : str
        Name of the target logical qubit.
    """

    target_qubit: str


# what about non-Clifford gates?
@dataclass(config=dataclass_config)
class T(LogicalOperation):
    """
    Describes a T gate on a logical qubit.

    Parameters
    ----------
    target_qubit : str
        Name of the target logical qubit.
    """

    target_qubit: str



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from functools import cached_property, partial
from typing import Any

from pydantic import Field


from ..eka import Circuit, Channel
from ..eka.utilities import BoolOp

from .op_signature import (
    ALL_EKA_OP_SIGNATURES,
    BOOL_LOGIC_OP_SIGNATURE,
    CLIFFORD_GATES_SIGNATURE,
    NONCLIFFORD_GATES_SIGNATURE,
    CONTROL_FLOW_OP_SIGNATURE,
    UTILS_SIGNATURE,
    OpType,
    OpSignature,
)
from .converter import Converter, OpAndTargetToInstrCallable


QasmProgram = tuple[str, dict[Channel, int], dict[Channel, int]]


class EkaToQasmConverter(Converter[QasmProgram, Any]):
    """Convert EKA circuits to OpenQASM3.0 format.
    This converter translates Eka operations into their corresponding OpenQASM3.0
    instructions, handling quantum and classical channels appropriately.

    Here's a simple example of how to use this method to execute Eka experiment in
    Qiskit through QASM conversion:

    .. code-block:: python

        from loom.executor import EkaToQasmConverter

        qasm_program, q_register, c_register = EkaToQasmConverter().convert(eka)

        from qiskit import QuantumCircuit, transpile
        from qiskit_aer import AerSimulator
        import qiskit.qasm3

        circuit = qiskit.qasm3.loads(qasm_program)
        simulator = AerSimulator()
        qc_t = transpile(circuit, simulator)

        result = simulator.run(qc_t, shots=5).result().get_counts()

        parsed_outcome = EkaToQasmConverter.parse_target_run_outcome(
            (result, c_register)
        )

    """

    SUPPORTED_OPERATIONS: frozenset[OpSignature] = ALL_EKA_OP_SIGNATURES

    ALLOW_ERROR_MODELS: bool = Field(default=False, frozen=True, init=False)

    @cached_property
    # pylint: disable=unused-argument
    def operations_map(self) -> dict[str, OpAndTargetToInstrCallable]:
        """Map of operation signatures to their corresponding QASM instructions."""
        eka_to_qasm_ops = {
            "i": "id",
            "x": "x",
            "y": "y",
            "z": "z",
            "h": "h",
            "t": "t",
            "phase": "s",
            "phaseinv": "sdg",
            "cnot": "cx",
            "cx": "cx",
            "cy": "cy",
            "cz": "cz",
            "swap": "swap",
            "reset": "reset",
            "reset_0": "reset",
            "reset_1": ["reset", "x"],
            "reset_+": ["reset", "h"],
            "reset_-": ["reset", "x", "h"],
            "reset_+i": ["reset", "h", "s"],
            "reset_-i": ["reset", "h", "sdg"],
            "measurement": "measure",
            "measure_z": "measure",
            "measure_x": ["h", "measure"],
            "measure_y": ["sdg", "h", "measure"],
        }

        def _single_qubit_op(
            q_target: list[str],
            c_target: list[str],
            op: str | list[str],
            desc: str = "",
        ) -> str:
            if not isinstance(op, list):
                op = [op]
            qasm_ops = []
            for o in op:
                qasm_ops.append(f"{o} {q_target[0]};")
            return "\n".join(qasm_ops)

        def _two_qubit_op(
            q_target: list[str],
            c_target: list[str],
            op: str,
            desc: str = "",
        ) -> str:
            return f"{op} {q_target[0]}, {q_target[1]};"

        def _measurement_op(
            q_target: list[str],
            c_target: list[str],
            op: str | list[str],
            desc: str = "",
        ) -> list[str]:
            qasm_ops = []
            if not isinstance(op, list):
                op = [op]
            for o in op[:-1]:
                qasm_ops.append(f"{o} {q_target[0]};")
            qasm_ops.append(f"{op[-1]} {q_target[0]} -> {c_target[0]};")
            return "\n".join(qasm_ops)

        def _control_flow_op(
            q_target: list[str], c_target: list[str], op: str, desc: str = ""
        ) -> str:
            """Handling control flow operations in QASM. c_target[0] is the classical
            condition instruction string already formatted."""
            match op:
                case "classical_if":
                    return f"if ({desc}) {{"
                case "classical_else":
                    return "} else {"
                case "end_if":
                    return "}"
                case _:
                    raise ValueError(f"Unsupported control flow operation: {op}")

        def _bool_logic_op(
            q_target: list[str], c_target: list[str], op: str, desc: str = ""
        ) -> str:
            """Handling boolean logic operations in QASM. c_target contains the
            classical channel targets."""
            condition = ""
            match op:
                case BoolOp.MATCH:
                    condition = c_target[0]
                case BoolOp.NOT:
                    condition = f"!{c_target[0]}"
                case BoolOp.AND:
                    condition = " == 1 && ".join(c_target) + " == 1"
                case BoolOp.OR:
                    condition = " == 1 || ".join(c_target) + " == 1"
                case BoolOp.XOR:
                    condition = " == 1 ^ ".join(c_target) + " == 1"
                case BoolOp.NAND:
                    condition = f"!({' == 1 && '.join(c_target)} == 1)"
                case BoolOp.NOR:
                    condition = f"!({' == 1 || '.join(c_target)} == 1)"
                case _:
                    condition = ValueError(f"Unsupported BoolOp '{op}'.")
            return condition

        def _utils_op(
            q_target: list[str], c_target: list[str], op: str, desc: str = ""
        ) -> str:
            """Handling utility operations in QASM."""
            match op:
                case "comment":
                    return f"// {desc}"
                case _:
                    return None

        # Map operation types to their corresponding function
        op_type_handlers = {
            OpType.SINGLE_QUBIT: _single_qubit_op,
            OpType.TWO_QUBIT: _two_qubit_op,
            OpType.MEASUREMENT: _measurement_op,
            OpType.RESET: _single_qubit_op,
            OpType.CONTROL_FLOW: _control_flow_op,
            OpType.BOOL_LOGIC: _bool_logic_op,
            OpType.UTILS: _utils_op,
        }

        return {
            # Quantum gate instructions
            op.name: partial(op_type_handlers[op.op_type], op=eka_to_qasm_ops[op.name])
            for op in CLIFFORD_GATES_SIGNATURE | NONCLIFFORD_GATES_SIGNATURE
        } | {
            # Bool op and control flow mapping are less straightforward so they are
            # handled in the handlers methods
            op.name: partial(op_type_handlers[op.op_type], op=op.name)
            for op in BOOL_LOGIC_OP_SIGNATURE
            | CONTROL_FLOW_OP_SIGNATURE
            | UTILS_SIGNATURE
        }

    def convert_circuit(self, input_circuit: Circuit) -> QasmProgram:
        """
        Convert an Eka circuit to OpenQASM3.0 format.

        Parameters
        ----------
        input_data : Circuit
            The Eka circuit to convert.

        Returns
        -------
        QasmProgram (tuple[str, dict[Channel, int], dict[Channel, int]])
            A tuple containing the QASM program as a string, a mapping from quantum
            channels to their indices, and a mapping from classical channels to their
            indices.
        """
        # List, where each element is a line in the QASM program
        qasm_str_program = ["OPENQASM 3.0;", 'include "stdgates.inc";']

        # pylint: disable=unsupported-membership-test
        if (
            not input_circuit.circuit and input_circuit.name not in self.operations_map
        ) or not input_circuit.channels:
            return "# empty input", {}, {}

        init_instructions, quant_ch_to_idx, classical_ch_to_idx = (
            self.emit_init_instructions(input_circuit)
        )
        qasm_str_program.append("// Init registers")
        qasm_str_program.extend(init_instructions.splitlines())

        circuit_instructions = self.emit_circuit_program(
            input_circuit, quant_ch_to_idx, classical_ch_to_idx
        )

        qasm_str_program.extend(circuit_instructions.splitlines())

        return (
            "\n".join(qasm_str_program),
            quant_ch_to_idx,
            classical_ch_to_idx,
        )

    @staticmethod
    def parse_target_run_outcome(
        run_output: tuple[dict[str, int], dict[Channel, int]],
    ) -> dict[str, int | list[int]]:
        """
        Parse the output of a target run by mapping the bitstring to the register
        labels.

        Parameters
        ----------
        run_output : tuple[dict[str, int], dict[Channel, int]]
            Output of the simulation run as a tuple of:

            - bitstrings: dict with keys as bitstrings and values as their counts,
            - channel_to_idx: a mapping from channels to their bitstring indices,
              this mapping is given by the converter when exporting the circuit.

        Returns
        -------
        dict[str, int | list[int]]
            dict mapping register labels to their corresponding values (a list where
            each element represents the value of the register in a specific shot).
        """
        bitstrings, channel_to_idx = run_output

        bitstrings = [key for key, count in bitstrings.items() for _ in range(count)]
        # Implement the parsing logic here
        result = {chan.label: [] for chan in channel_to_idx.keys()}

        if bitstrings == []:
            raise ValueError("No bitstrings found in the run output.")
        num_shots = len(bitstrings[0])

        for channel in channel_to_idx.keys():
            idx = channel_to_idx[channel]
            # Qiskit returns bitstrings in reverse order
            reversed_index = num_shots - 1 - idx
            result[channel.label] = [int(b[reversed_index]) for b in bitstrings]

        return result

    def emit_init_instructions(
        self, circuit: Circuit
    ) -> tuple[str, dict[Channel, Any], dict[Channel, Any]]:
        """Provide the python code to initializes the
        quantum and classical registers, and return the mapping from eka channel to
        register index."""

        if not isinstance(circuit, Circuit):
            raise TypeError("Input must be a Circuit instance.")

        if circuit.circuit == [] and circuit.channels == []:
            return "", {}, {}

        q_channels = sorted(
            [channel for channel in circuit.channels if channel.is_quantum()],
            key=lambda ch: ch.label,
        )

        c_channels = sorted(
            [channel for channel in circuit.channels if channel.is_classical()],
            key=lambda ch: ch.label,
        )

        qasm_str_program = []
        # Init registers
        if len(q_channels) > 0:
            qasm_str_program.append(f"qubit[{len(q_channels)}] q;")
        if len(c_channels) > 0:
            qasm_str_program.append(f"bit[{len(c_channels)}] c;")

        quant_ch_to_idx = {ch: idx for idx, ch in enumerate(q_channels)}
        classical_ch_to_idx = {ch: idx for idx, ch in enumerate(c_channels)}

        return "\n".join(qasm_str_program), quant_ch_to_idx, classical_ch_to_idx

    def emit_leaf_circuit_instruction(
        self,
        circuit: Circuit,
        quantum_channel_map: dict[Channel, int],
        classical_channel_map: dict[Channel, int],
    ) -> str:
        """Provide the python code to emit an Eka instruction in the target
        language."""

        if not isinstance(circuit, Circuit):
            raise TypeError("Input must be a Circuit instance.")

        if circuit.circuit and len(circuit.circuit) != 0:
            raise ValueError("The circuit must be an leaf circuit with channels.")

        op_name = circuit.name

        q_target = [
            f"q[{quantum_channel_map[qt]}]"
            for qt in circuit.channels
            if qt.is_quantum()
        ]

        c_target = [
            f"c[{classical_channel_map[ct]}]"
            for ct in circuit.channels
            if ct.is_classical()
        ]

        self._validate_ops_args(op_name, len(q_target), len(c_target))

        # pylint: disable=unsupported-membership-test
        if op_name not in self.operations_map:
            raise KeyError(f"Operation {op_name} not supported in QASM target.")

        return self.operations_map[op_name](
            q_target, c_target, desc=circuit.description
        )
